//-----------------------------------------------------------------------
//  Java Indexed Serialization Package
//  ----------------------------------
//  BTreeIndex.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.text.*;
import java.io.*;
import java.util.*;

/**
 * Associates a <i>key</i> value with a <code>long</code> reference. In general, the
 * reference is a file pointer that locates a serializable object associated with
 * the key value; however, the index itself does not apply semantics to the reference.
 * The user of index interprets the reference according to its design; an
 * <code>IndexedObjectDatabase</code>, for example, uses the reference to record the
 * file position of a <code>Serializable</code> object with a specific key.
 * <p>
 * Use the <code>IndexedObjectDatabase.attachIndex</code> method to attach a
 * <code>BTreeIndex</code> to a database. A database updates all attached indexes
 * when it writes or removes objects from its file. This is the most common use
 * of <code>BTreeIndex</code>es.
 * <p>
 * It's important to realize that each key is associated with a <code>long</code>
 * <i>reference</i> which is interpretted by the user. For example, the <code>
 * IndexedObjectDatabase</code> class stores a file pointer in reference, thus
 * associating a key value with a serialized object stored in the database. You can
 * also use the reference as an index into an array, or in any other way appropriate
 * to your application. It is even possible associate the same reference value with
 * different keys.
 * <p>
 *
 * @author  Scott Robert Ladd
 * @see		com.coyotegulch.jisp.ObjectIndex
 * @see		com.coyotegulch.jisp.IndexedObjectDatabase
 */
public class BTreeIndex extends BTreePageFile implements ObjectIndex
{
    // exceptions
    private static final KeyNotFound  err_key_not_found = new KeyNotFound("key not found in BTreeIndex");
    private static final DuplicateKey err_duplicate_key = new DuplicateKey("duplicate key found in BTreeIndex");
    
    // inner classes
    private class SearchResult
    {
        public boolean m_found;
        public Page    m_page;
        public int     m_position;

        // constructor
        public SearchResult(boolean found, Page page, int  position)
        {
            m_found    = found;
            m_page     = page;
            m_position = position;
        }
    }
    
    // working storage
    private Page m_root;
    private IndexedObjectDatabase m_database = null;
    
    /**
     * Calculates a suggested value for a B-tree order, based on a given number of records and a maximum
     * "height" for the B-tree structure.
     * 
     * @param   numRecords  number of records expected in a B-tree index
     * @param   maxHeight   maximum allowable height for the B-tree (i.e., maximum number of pages to be read
     *                      when searching for a key)
     * @return  a suggested value for oder, to be used in constructing a new BTreeIndex
     * @see     com.coyotegulch.jisp.BTreeIndex
     */
    public static int computeOrder(int numRecords, int maxHeight)
    {
        int result;
        
        if ((numRecords < 1) || (maxHeight < 1))
            result = 7;
        else
            result = (int)Math.pow((double)numRecords,1.0 / (double)(maxHeight)) + 1;
            
        return result;
    }
    
    /**
     * Opens an existing file, <code>name</code>, as a <code>BTreeIndex</code>.
     *
     * @param   name    name of the external index file to be opened.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     */    
    public BTreeIndex(String name) throws IOException, ClassNotFoundException
    {
        // invoke superclass constructor
        super(name);

        // read root page of index
        m_root = readRoot();
    }

    /**
     * Creates the specified file as a <code>BTreeIndex</code>, using the given order and
     * <code>null</code> key value. The null key is a "blank" key used in padding empty
     * entries in B-tree pages. Obtain a null key for any key type by either invoking the
     * default constructor or calling the <code>KeyObject.makeNullKey</code> method.
     *
     * @param   name        name of the external index file to be opened.
     * @param   order       order (page size) the the B-Tree. This should be an odd number
     *                      greater than or equal to five (5). <code>BTreeIndex</code> will
     *                      increment any even numbered order to the next consecutive odd number
     *                      (if you specfy an order of 32, the index will actually have an order
     *                      of 33). Also, an order less than 5 will be set to 5 automatically.
     * @param   nullKey                 identifies a blank table entry and provides type-checking on new
     *                                  keys added to the index. All keys added to the index must have the
     *                                  same type as <code>nullkey</code>.
     * @param   hasDupes                Determines whether or not this index allows duplicate keys.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @see     com.coyotegulch.jisp.KeyObject
     */    
    public BTreeIndex(String name, int order, KeyObject nullKey, boolean hasDupes) throws IOException, ClassNotFoundException
    {
        // invoke superclass constructor
        super(name,order,nullKey,hasDupes);

        // read root page of index
        m_root = readRoot();
    }

    /**
     *  Closes an open <code>BTreeIndex</code> and empties the page cache to release memory.
     */
    public void close() throws IOException
    {
        emptyPageCache();
        super.close();
    }
    
    /**
     *  Returns the number of keys stored in this index. It is incremented when a new
     *  key is added, and decremented when a key is removed.
     *
     *  @return the number of keys stored in this index
     */
    public int count()
    {
        return m_header.m_count;
    }

    /**
     *  Returns the number of keys added to this index since its creation. This value is
     *  incremented when a new key is added, but (unlike <code>count</count>, it is never
     *  decremented. In an index with a one-to-one correspondence between records and keys,
     *  this value provides a unique "record ID".
     *
     *  @return the number of keys added to this index since its creation
     */
    public long ticker()
    {
        return m_header.m_ticker;
    }

    /**
     * Insert a key into the index, associating a "reference" value with the given key.
     *
     * @param   key                     identifies the record to be read
     * @param   reference               reference associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DuplicateKey            when inserting a duplicate key into an index that does not
     *                                  support duplicates
     * @see     com.coyotegulch.jisp.DuplicateKey
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void insertKey(KeyObject key, long reference) throws IOException, ClassNotFoundException
    {
        // search for the key
        SearchResult ins = searchInsert(m_root,key);

        if (ins.m_found && !m_header.m_hasDupes)
            throw err_duplicate_key;
        else
        {
            writeKey(ins,key,reference);
            ++m_header.m_count;
            ++m_header.m_ticker;
            rewriteHeader();
        }
    }
    
    /**
     * Replaces the reference for the given key. The method replaces the <i>first</i> occurrence
     * of the key, if the index includes duplicates.
     *
     * @param   key                     identifies the record being replaced
     * @param   reference               new reference to be associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.DuplicateKey
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void replaceKey(KeyObject key, long reference) throws IOException, ClassNotFoundException
    {
        // search for the key
        SearchResult ins = search(m_root,key,false);

        if (ins.m_found)
        {
            // store new data record pointer
            ins.m_page.m_recPtr[ins.m_position] = reference;

            // rewrite modified page
            write(ins.m_page,false);
        }
        else
        {
            throw err_key_not_found;
        }

        m_root = readRoot();
    }
    
    /**
     * Replaces or inserts the reference for the given key. If the key exists, the reference
     * associated with that key is replaced; if the key does not exist, a new key is
     * inserted into the database with the given reference. If the index supports duplicate keys,
     * the first-found occurence of the key will be replaced.
     *
     * @param   key                     identifies the record to be stored
     * @param   position                reference associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void storeKey(KeyObject key, long reference) throws IOException, ClassNotFoundException
    {
        // search for the key
        SearchResult ins = search(m_root,key,false);

        if (ins.m_found)
        {
            // store new data record pointer
            ins.m_page.m_recPtr[ins.m_position] = reference;

            // rewrite modified page
            write(ins.m_page,false);
        }
        else
        {
            writeKey(ins,key,reference);
            ++m_header.m_count;
            ++m_header.m_ticker;
            rewriteHeader();
        }
    }
    
    /**
     * Find the position of the object associated with a given key.
     *
     * @param   key                     a key to be sought in the index
     * @return                          Position of the record associated with <code>key</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public long findKey(KeyObject key) throws IOException, ClassNotFoundException
    {
        return findKey(key,false);
    }
    
    /**
     * Find the position of an object associated with a given key, or its successor.
     *
     * @param   key                     a key to be sought in the index
     * @param   allowNext               when <code>true</code>, <code>findKey</code> will return
     *                                  the reference associated with the key greater than or
     *                                  equal to <code>key</code>
     * @return                          The reference associated with <code>key</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public long findKey(KeyObject key, boolean allowNext) throws IOException, ClassNotFoundException
    {
        // search for key
        SearchResult result = search(m_root,key,allowNext);

        if (result.m_found)
            return result.m_page.m_recPtr[result.m_position];
        else
            throw err_key_not_found;
    }
    
    /**
     * Removes the given key from the index. If the index contains duplicates of <code>key</code>,
     * the first key found will be deleted.
     *
     * @param   key                     A key identifying the record to be read.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void removeKey(KeyObject key) throws IOException, ClassNotFoundException
    {
        // search for key
        SearchResult del = search(m_root,key,false);

        if (del.m_found == false)
            throw err_key_not_found;

        // is this a leaf node?
        if (del.m_page.m_link[0] == NULL_PTR)
        {
            // remove key from leaf
            --del.m_page.m_header.m_numKeys;

            // slide keys left over deleted one
            for (int n = del.m_position; n < del.m_page.m_header.m_numKeys; ++n)
            {
                del.m_page.m_key[n] = del.m_page.m_key[n + 1];
                del.m_page.m_recPtr[n] = del.m_page.m_recPtr[n + 1];
            }

            del.m_page.m_key[del.m_page.m_header.m_numKeys] = m_header.m_nullKey.makeNullKey();
            del.m_page.m_recPtr[del.m_page.m_header.m_numKeys] = NULL_PTR;

            // write changed record
            write(del.m_page,false);

            // adjust the tree, if needed
            if (del.m_page.m_header.m_numKeys < m_header.m_minKeys)
                adjustTree(del.m_page);
        }
        else // deleting an internal page
        {
            // get the successor page
            Page sucPage = read(del.m_page.m_link[del.m_position + 1]);

            while (sucPage.m_link[0] != NULL_PTR)
                sucPage = read(sucPage.m_link[0]);

            // first key is the "swappee"
            del.m_page.m_key[del.m_position] = sucPage.m_key[0];
            del.m_page.m_recPtr[del.m_position] = sucPage.m_recPtr[0];

            // delete swapped key from successor page
            --sucPage.m_header.m_numKeys;

            for (int n = 0; n < sucPage.m_header.m_numKeys; ++n)
            {
                sucPage.m_key[n]      = sucPage.m_key[n + 1];
                sucPage.m_recPtr[n]   = sucPage.m_recPtr[n + 1];
                sucPage.m_link[n + 1] = sucPage.m_link[n + 2];
            }

            sucPage.m_key[sucPage.m_header.m_numKeys]      = m_header.m_nullKey.makeNullKey();
            sucPage.m_recPtr[sucPage.m_header.m_numKeys]   = NULL_PTR;
            sucPage.m_link[sucPage.m_header.m_numKeys + 1] = NULL_PTR;

            // write modified records
            write(del.m_page,(del.m_page.m_header.m_parentPtr == NULL_PTR)); // false);
            write(sucPage,false);

            // adjust tree for leaf node
            if (sucPage.m_header.m_numKeys < m_header.m_minKeys)
                adjustTree(sucPage);
        }
        
        --m_header.m_count;
        rewriteHeader();
    }
    
    // find a node
    private SearchResult search(Page pg, KeyObject key, boolean allowNext) throws IOException, BTreeException, ClassNotFoundException
    {
        int position = 0;

        // if page is empty, we didn't find the key
        if (pg.m_header.m_numKeys == 0)
            return new SearchResult(false,pg,position);

        // loop through keys
        while (true)
        {
            // watch for end of page
            if (position == pg.m_header.m_numKeys)
                break;

            // compare keys
            int comp = pg.m_key[position].compareTo(key);

            // return result if key was found
            if (comp == KeyObject.KEY_EQUAL || (allowNext && (comp == KeyObject.KEY_MORE)))
                return new SearchResult(true,pg,position);
            else
            {
                // otherwise, continue searching
                if (comp == KeyObject.KEY_LESS)
                    ++position; // this page
                else
                    break; // search successor page
            }
        }

        // if we're in a leaf, the key hasn't been found
        if (pg.m_link[position] == NULL_PTR)
            return new SearchResult(false,pg,position);
        else // read and search next page
            return search(read(pg.m_link[position]),key,allowNext);
    }

    // find a node
    private SearchResult searchWithPosition(Page pg, KeyObject key, boolean allowNext) throws IOException, BTreeException, ClassNotFoundException
    {
        int position = 0;

        // if page is empty, we didn't find the key
        if (pg.m_header.m_numKeys == 0)
            return new SearchResult(false,pg,position);

        // loop through keys
        while (true)
        {
            // watch for end of page
            if (position == pg.m_header.m_numKeys)
                break;

            // compare keys
            int comp = pg.m_key[position].compareTo(key);
            
            // return result if key was found
            if (comp == KeyObject.KEY_EQUAL || (allowNext && (comp == KeyObject.KEY_MORE)))
                return new SearchResult(true,pg,position);
            else
            {
                // otherwise, continue searching
                if (comp == KeyObject.KEY_LESS)
                    ++position; // this page
                else
                    break; // search successor page
            }
        }

        // if we're in a leaf, the key hasn't been found
        if (pg.m_link[position] == NULL_PTR)
            return new SearchResult(false,pg,position);
        else // read and search next page
            return search(read(pg.m_link[position]),key,allowNext);
    }

    // search for leaf node in which we insert the given key    
    private SearchResult searchInsert(Page pg, KeyObject key) throws IOException, BTreeException, ClassNotFoundException
    {
        int position = 0;

        // if page is empty, we didn't find the key
        if (pg.m_header.m_numKeys == 0)
            return new SearchResult(false,pg,position);

        // loop through keys
        while (true)
        {
            // watch for end of page
            if (position == pg.m_header.m_numKeys)
                break;

            // compare keys
            int comp = pg.m_key[position].compareTo(key);
            
            if (comp == KeyObject.KEY_LESS)
                ++position;
            else
                break;
        }

        // if we're in a leaf, the key hasn't been found
        if (pg.m_link[position] == NULL_PTR)
            return new SearchResult(false,pg,position);
        else // read and search next page
            return searchInsert(read(pg.m_link[position]),key);
    }
    

    // internal function to write keys
    private void writeKey(SearchResult ins, KeyObject key, long position) throws IOException, ClassNotFoundException
    {
        // check to see if page is full
        if (ins.m_page.m_header.m_numKeys == m_header.m_maxKeys)
        {
            // temporary arrays
            KeyObject [] tempKeys = new KeyObject [m_header.m_maxKeys + 1];
            long      [] tempPtrs = new long [m_header.m_maxKeys + 1];

            // store new item
            tempKeys[ins.m_position] = key;
            tempPtrs[ins.m_position] = position;

            // copy entries from insertion page to temps
            int nt = 0; // index for temp arrays
            int ni = 0; // index for insertion page

            while (ni < m_header.m_maxKeys)
            {
                // skip over inserted data
                if (ni == ins.m_position)
                    ++nt;

                // copy data
                tempKeys[nt] = ins.m_page.m_key[ni];
                tempPtrs[nt] = ins.m_page.m_recPtr[ni];

                // next one
                ++ni;
                ++nt;
            }

            // create a new leaf
            Page sibPage = new Page(m_header);
            sibPage.m_header.m_parentPtr = ins.m_page.m_header.m_parentPtr;

            // clear key counts
            ins.m_page.m_header.m_numKeys = 0;
            sibPage.m_header.m_numKeys    = 0;

            // copy keys from temp to pages
            for (ni = 0; ni < m_header.m_minKeys; ++ni)
            {
                ins.m_page.m_key[ni]    = tempKeys[ni];
                ins.m_page.m_recPtr[ni] = tempPtrs[ni];

                ++ins.m_page.m_header.m_numKeys;
            }

            for (ni = m_header.m_minKeys + 1; ni <= m_header.m_maxKeys; ++ni)
            {
                sibPage.m_key[ni - 1 - m_header.m_minKeys]    = tempKeys[ni];
                sibPage.m_recPtr[ni - 1 - m_header.m_minKeys] = tempPtrs[ni];

                ++sibPage.m_header.m_numKeys;
            }

            // replace remaining entries with null
            for (ni = m_header.m_minKeys; ni < m_header.m_maxKeys; ++ni)
            {
                ins.m_page.m_key[ni]    = m_header.m_nullKey.makeNullKey();
                ins.m_page.m_recPtr[ni] = NULL_PTR;
            }

            // write pages
            write(ins.m_page,false);
            write(sibPage,false);

            // promote key and its pointer
            if (ins.m_page.m_header.m_parentPtr == NULL_PTR)
            {
                // creating a new m_root page
                promoteRoot(tempKeys[m_header.m_minKeys],
                            tempPtrs[m_header.m_minKeys],
                            ins.m_page,
                            sibPage);
            }
            else
            {
                // promote key into parent page
                Page parPage = read(ins.m_page.m_header.m_parentPtr);

                promoteInternal(parPage,
                                tempKeys[m_header.m_minKeys],
                                tempPtrs[m_header.m_minKeys],
                                sibPage.m_header.m_filePtr);
            }
        }
        else
        {
            // move keys to make room for new one
            for (int n = ins.m_page.m_header.m_numKeys; n > ins.m_position; --n)
            {
                ins.m_page.m_key[n]    = ins.m_page.m_key[n - 1];
                ins.m_page.m_recPtr[n] = ins.m_page.m_recPtr[n - 1];
            }

            ins.m_page.m_key[ins.m_position]    = key;
            ins.m_page.m_recPtr[ins.m_position] = position;

            ++ins.m_page.m_header.m_numKeys;

            // write updated page
            write(ins.m_page,false);
        }

        m_root = readRoot();
    }

    // promote key into parent
    private void promoteInternal(Page insPage, KeyObject key, long ptr, long link) throws IOException, BTreeException, ClassNotFoundException
    {
        if (insPage.m_header.m_numKeys == m_header.m_maxKeys)
        {
            // temporary array
            KeyObject [] tempKeys = new KeyObject [m_header.m_maxKeys + 1];
            long []      tempPtrs = new long      [m_header.m_maxKeys + 1];
            long []      tempLnks = new long      [m_header.m_order   + 1];

            tempLnks[0] = insPage.m_link[0];

            // copy entries from inapage to temps
            int nt  = 0;
            int ni  = 0;
            int position = 0;

            // find insertion point
            while ((position < insPage.m_header.m_numKeys) && (insPage.m_key[position].compareTo(key) == KeyObject.KEY_LESS))
                ++position;

            // store new info
            tempKeys[position]     = key;
            tempPtrs[position]     = ptr;
            tempLnks[position + 1] = link;

            // copy existing keys
            while (ni < m_header.m_maxKeys)
            {
                if (ni == position)
                    ++nt;

                tempKeys[nt]     = insPage.m_key[ni];
                tempPtrs[nt]     = insPage.m_recPtr[ni];
                tempLnks[nt + 1] = insPage.m_link[ni + 1];

                ++ni;
                ++nt;
            }

            // generate a new leaf node
            Page sibPage = new Page(m_header);

            sibPage.m_header.m_parentPtr = insPage.m_header.m_parentPtr;

            // clear key counts
            insPage.m_header.m_numKeys = 0;
            sibPage.m_header.m_numKeys = 0;

            insPage.m_link[0] = tempLnks[0];

            // copy keys from temp to pages
            for (ni = 0; ni < m_header.m_minKeys; ++ni)
            {
                insPage.m_key[ni]      = tempKeys[ni];
                insPage.m_recPtr[ni]   = tempPtrs[ni];
                insPage.m_link[ni + 1] = tempLnks[ni + 1];

                ++insPage.m_header.m_numKeys;
            }

            sibPage.m_link[0] = tempLnks[m_header.m_minKeys + 1];

            for (ni = m_header.m_minKeys + 1; ni <= m_header.m_maxKeys; ++ni)
            {
                sibPage.m_key[ni - 1 - m_header.m_minKeys]    = tempKeys[ni];
                sibPage.m_recPtr[ni - 1 - m_header.m_minKeys] = tempPtrs[ni];
                sibPage.m_link[ni - m_header.m_minKeys]       = tempLnks[ni + 1];

                ++sibPage.m_header.m_numKeys;
            }

            // replace remaining entries with null
            for (ni = m_header.m_minKeys; ni < m_header.m_maxKeys; ++ni)
            {
                insPage.m_key[ni]      = m_header.m_nullKey.makeNullKey();
                insPage.m_recPtr[ni]   = NULL_PTR;
                insPage.m_link[ni + 1] = NULL_PTR;
            }

            // write pages
            write(insPage,false);
            write(sibPage,false);

            // update parent links in child nodes
            for (ni = 0; ni <= sibPage.m_header.m_numKeys; ++ni)
            {
                Page child = read(sibPage.m_link[ni]);
                child.m_header.m_parentPtr = sibPage.m_header.m_filePtr;
                write(child,false);
            }

            // promote key and pointer
            if (insPage.m_header.m_parentPtr == NULL_PTR)
            {
                // create a new m_root
                promoteRoot(tempKeys[m_header.m_minKeys],
                            tempPtrs[m_header.m_minKeys],
                            insPage,
                            sibPage);
            }
            else
            {
                // read parent and promote key
                Page parPage = read(insPage.m_header.m_parentPtr);

                promoteInternal(parPage,
                                tempKeys[m_header.m_minKeys],
                                tempPtrs[m_header.m_minKeys],
                                sibPage.m_header.m_filePtr);
            }
        }
        else
        {
            int position = 0;

            // find insertion point
            while ((position < insPage.m_header.m_numKeys) && (insPage.m_key[position].compareTo(key) == KeyObject.KEY_LESS))
                ++position;

            // shift keys right
            for (int n = insPage.m_header.m_numKeys; n > position; --n)
            {
                insPage.m_key[n]      = insPage.m_key[n - 1];
                insPage.m_recPtr[n]   = insPage.m_recPtr[n - 1];
                insPage.m_link[n + 1] = insPage.m_link[n];
            }

            insPage.m_key[position]      = key;
            insPage.m_recPtr[position]   = ptr;
            insPage.m_link[position + 1] = link;

            ++insPage.m_header.m_numKeys;

            // write updated page
            write(insPage,false);
        }
    }

    // promote key by creating new m_root
    private void promoteRoot(KeyObject key, long ptr, Page lessPage, Page grtrPage) throws IOException, BTreeException, ClassNotFoundException
    {
        // create a new m_root page
        Page newRoot = new Page(m_header);

        // add key and links to m_root
        newRoot.m_key[0]    = key;
        newRoot.m_recPtr[0] = ptr;
        newRoot.m_link[0]   = lessPage.m_header.m_filePtr;
        newRoot.m_link[1]   = grtrPage.m_header.m_filePtr;

        newRoot.m_header.m_numKeys = 1;

        // write new m_root to tree
        write(newRoot,true);

        // update children
        lessPage.m_header.m_parentPtr = newRoot.m_header.m_filePtr;
        grtrPage.m_header.m_parentPtr = newRoot.m_header.m_filePtr;

        write(lessPage,false);
        write(grtrPage,false);
    }

    // adjust tree if page is too small
    private void adjustTree(Page pg) throws IOException, BTreeException, ClassNotFoundException
    {
        // check for root
        if (pg.m_header.m_parentPtr == NULL_PTR)
            return;

        // get parent page
        Page parPage = read(pg.m_header.m_parentPtr);

        // find pointer to pg
        int n = 0;

        while (parPage.m_link[n] != pg.m_header.m_filePtr)
            ++n;

        // read sibling pages
        Page sibMore = null;
        Page sibLess = null;

        if (n < parPage.m_header.m_numKeys)
            sibMore = read(parPage.m_link[n+1]);
 
        if (n > 0)
            sibLess = read(parPage.m_link[n-1]);
        
        // figure out what to do
        if (sibLess != null)
        {
            --n;

            if (sibLess.m_header.m_numKeys > m_header.m_minKeys)
                redistribute(n,sibLess,parPage,pg);
            else
                concatenate(n,sibLess,parPage,pg);
        }
        else
        {
            if (sibMore != null)
            {
                if (sibMore.m_header.m_numKeys > m_header.m_minKeys)
                    redistribute(n,pg,parPage,sibMore);
                else
                    concatenate(n,pg,parPage,sibMore);
            }
            else
                throw new BTreeException("This should never happen!");
        }
    }

    // redistribute keys among siblings and parent
    private void redistribute(int position, Page lessPage, Page parPage, Page morePage) throws IOException, BTreeException, ClassNotFoundException
    {
        // check for leaf page
        if (lessPage.m_link[0] == NULL_PTR)
        {
            if (lessPage.m_header.m_numKeys > morePage.m_header.m_numKeys)
            {
                // move a key from lessPage to morePage
                for (int n = morePage.m_header.m_numKeys; n > 0; --n)
                {
                    morePage.m_key[n]    = morePage.m_key[n - 1];
                    morePage.m_recPtr[n] = morePage.m_recPtr[n - 1];
                }

                // store parent separator in morePage
                morePage.m_key[0]    = parPage.m_key[position];
                morePage.m_recPtr[0] = parPage.m_recPtr[position];

                // increment morePage key count
                ++morePage.m_header.m_numKeys;

                // decrement lessPage key count
                --lessPage.m_header.m_numKeys;

                // move last key in lessPage to parPage as separator
                parPage.m_key[position]    = lessPage.m_key[lessPage.m_header.m_numKeys];
                parPage.m_recPtr[position] = lessPage.m_recPtr[lessPage.m_header.m_numKeys];

                // clear last key in lessPage
                lessPage.m_key[lessPage.m_header.m_numKeys]    = m_header.m_nullKey.makeNullKey();
                lessPage.m_recPtr[lessPage.m_header.m_numKeys] = NULL_PTR;
            }
            else
            {
                // add parent key to lesser page
                lessPage.m_key[lessPage.m_header.m_numKeys]    = parPage.m_key[position];
                lessPage.m_recPtr[lessPage.m_header.m_numKeys] = parPage.m_recPtr[position];
                
                // increment lessPage key count
                ++lessPage.m_header.m_numKeys;

                // move first key in morePage to parPage as separator
                parPage.m_key[position]    = morePage.m_key[0];
                parPage.m_recPtr[position] = morePage.m_recPtr[0];

                // decrement morePage key count
                --morePage.m_header.m_numKeys;

                // move a key from morePage to lessPage
                int n;
                
                for (n = 0; n < morePage.m_header.m_numKeys; ++n)
                {
                    morePage.m_key[n]    = morePage.m_key[n + 1];
                    morePage.m_recPtr[n] = morePage.m_recPtr[n + 1];
                }

                // clear last key in morePage
                morePage.m_key[n]    = m_header.m_nullKey.makeNullKey();
                morePage.m_recPtr[n] = NULL_PTR;
            }
        }
        else
        {
            if (lessPage.m_header.m_numKeys > morePage.m_header.m_numKeys)
            {
                // move a key from lessPage to morePage
                for (int n = morePage.m_header.m_numKeys; n > 0; --n)
                {
                    morePage.m_key[n]      = morePage.m_key[n - 1];
                    morePage.m_recPtr[n]   = morePage.m_recPtr[n - 1];
                    morePage.m_link[n + 1] = morePage.m_link[n];
                }

                morePage.m_link[1] = morePage.m_link[0];

                // store parPage separator key in morePage
                morePage.m_key[0]    = parPage.m_key[position];
                morePage.m_recPtr[0] = parPage.m_recPtr[position];
                morePage.m_link[0]   = lessPage.m_link[lessPage.m_header.m_numKeys];

                // update child link
                Page child = read(morePage.m_link[0]);

                child.m_header.m_parentPtr = morePage.m_header.m_filePtr;

                write(child,false);

                // increment morePage key count
                ++morePage.m_header.m_numKeys;

                // decrement lessPage key count
                --lessPage.m_header.m_numKeys;

                // move last key in lessPage to parPage as separator
                parPage.m_key[position]    = lessPage.m_key[lessPage.m_header.m_numKeys];
                parPage.m_recPtr[position] = lessPage.m_recPtr[lessPage.m_header.m_numKeys];

                // clear last key in lessPage
                lessPage.m_key[lessPage.m_header.m_numKeys]      = m_header.m_nullKey.makeNullKey();
                lessPage.m_recPtr[lessPage.m_header.m_numKeys]   = NULL_PTR;
                lessPage.m_link[lessPage.m_header.m_numKeys + 1] = NULL_PTR;
            }
            else
            {
                // store parPage separator key in lessPage
                lessPage.m_key[lessPage.m_header.m_numKeys]      = parPage.m_key[position];
                lessPage.m_recPtr[lessPage.m_header.m_numKeys]   = parPage.m_recPtr[position];
                lessPage.m_link[lessPage.m_header.m_numKeys + 1] = morePage.m_link[0];

                // update child link
                Page child = read(morePage.m_link[0]);

                child.m_header.m_parentPtr = lessPage.m_header.m_filePtr;

                write(child,false);

                // increment lessPage key count
                ++lessPage.m_header.m_numKeys;

                // move last key in morePage to parPage as separator
                parPage.m_key[position]    = morePage.m_key[0];
                parPage.m_recPtr[position] = morePage.m_recPtr[0];

                // decrement morePage key count
                --morePage.m_header.m_numKeys;

                // move a key from morePage to lessPage
                int n;

                for (n = 0; n < morePage.m_header.m_numKeys; ++n)
                {
                    morePage.m_key[n]    = morePage.m_key[n + 1];
                    morePage.m_recPtr[n] = morePage.m_recPtr[n + 1];
                    morePage.m_link[n]   = morePage.m_link[n + 1];
                }

                morePage.m_link[n] = morePage.m_link[n + 1];

                // clear last key in morePage
                morePage.m_key[n]      = m_header.m_nullKey.makeNullKey();
                morePage.m_recPtr[n]   = NULL_PTR;
                morePage.m_link[n + 1] = NULL_PTR;
            }
        }

        // write updated pages
        write(lessPage,false);
        write(parPage,false);
        write(morePage,false);

        // update cached m_root if necessary
        if (parPage.m_header.m_parentPtr == NULL_PTR)
            m_root = parPage;
    }

    // concatenate sibling pages
    private void concatenate(int position, Page lessPage, Page parPage, Page morePage) throws IOException, BTreeException, ClassNotFoundException
    {
        // move separator key from parPage into lessPage
        lessPage.m_key[lessPage.m_header.m_numKeys]      = parPage.m_key[position];
        lessPage.m_recPtr[lessPage.m_header.m_numKeys]   = parPage.m_recPtr[position];
        lessPage.m_link[lessPage.m_header.m_numKeys + 1] = morePage.m_link[0];

        // increment lessPage key count
        ++lessPage.m_header.m_numKeys;

        // delete separator from parPage
        --parPage.m_header.m_numKeys;

        int n;

        for (n = position; n < parPage.m_header.m_numKeys; ++n)
        {
            parPage.m_key[n]      = parPage.m_key[n + 1];
            parPage.m_recPtr[n]   = parPage.m_recPtr[n + 1];
            parPage.m_link[n + 1] = parPage.m_link[n + 2];
        }

        // clear unsused key from parent
        parPage.m_key[n]      = m_header.m_nullKey.makeNullKey();
        parPage.m_recPtr[n]   = NULL_PTR;
        parPage.m_link[n + 1] = NULL_PTR;

        // copy keys from morePage to lessPage
        int ng = 0;
		n = lessPage.m_header.m_numKeys;

		// combine pages
		while (ng < morePage.m_header.m_numKeys)
		{
		    ++lessPage.m_header.m_numKeys;

		    lessPage.m_key[n]      = morePage.m_key[ng];
		    lessPage.m_recPtr[n]   = morePage.m_recPtr[ng];
		    lessPage.m_link[n + 1] = morePage.m_link[ng + 1];

		    ++ng;
		    ++n;
		}

		// delete morePage
		delete(morePage);

        // is this a leaf?
        if (lessPage.m_link[0] != NULL_PTR)
        {
            // adjust child pointers
            for (n = 0; n <= lessPage.m_header.m_numKeys; ++n)
            {
                Page child = read(lessPage.m_link[n]);

                child.m_header.m_parentPtr = lessPage.m_header.m_filePtr;

                write(child,false);
            }
        }

        // write lessPage and parent
        if (parPage.m_header.m_numKeys == 0)
        {
            delete(parPage);

            lessPage.m_header.m_parentPtr = NULL_PTR;

            write(lessPage,true);

            m_root = lessPage;
        }
        else
        {
            write(parPage,false);
            write(lessPage,false);

            // reset cached m_root page if needed
            if (parPage.m_header.m_parentPtr == NULL_PTR)
                m_root = parPage;

            // if parent is too small, adjust
            if (parPage.m_header.m_numKeys < m_header.m_minKeys)
                adjustTree(parPage);
        }
    }
    
    /**
     * Empty the page cache.
     */
    public void emptyPageCache()
    {
        m_pageCache = new HashMap();
    }
    
    /**
     * Get the size of the page cache, in number of B-tree pages stored in memory.
     */
    public int getPageCacheSize()
    {
        return m_pageCache.size();
    }
    
    /**
     * Dumps the entire B-Tree to <code>System.out</code> for analysis. For diagnostic purposes only.
     * 
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DuplicateKey            when inserting a duplicate key into an index that does not
     *                                  support duplicates
     */
    public void dumpTree(PrintStream output) throws IOException, DuplicateKey, ClassNotFoundException
    {
        // recursively dump from the root
        output.println("ROOT: " + m_root.m_header.m_filePtr);
        dumpTreeRecursive(m_root,output);
    }
    
    private void dumpTreeRecursive(Page pg, PrintStream output) throws IOException, DuplicateKey, ClassNotFoundException
    {
        // create a formatter for numbers
        DecimalFormat numform1 = new DecimalFormat("#######0");
        DecimalFormat numform2 = new DecimalFormat("000");
        
        output.println();
        output.println("PAGE:" + numform1.format(pg.m_header.m_filePtr));
        
        for (int k = 0; k < pg.m_header.m_numKeys; ++k)
        {
            output.println("" + numform2.format(k) + ": "  + numform1.format(pg.m_link[k]));
            output.println("     " + pg.m_key[k].toString().trim());
        }
        
        // display last link
        output.println("" + numform2.format(pg.m_header.m_numKeys) + ": " + numform1.format(pg.m_link[pg.m_header.m_numKeys]));
        
        for (int n = 0; n < pg.m_header.m_numKeys + 1; ++n)
        {
            if (pg.m_link[n] >= 0)
                dumpTreeRecursive(read(pg.m_link[n]),output);
        }
    }
}
