//-----------------------------------------------------------------------
//  Java Indexed Serialization Package
//  ----------------------------------
//  BTreeIterator.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.IOException;
import java.util.Stack;

/**
 * <code>BTreeIterator</code> defines an object that references a specific object relative to other 
 * objects in an <code>ObjectIndex</code>. In many ways, a <code>BTreeIterator</code> is analogous
 * to the "cursors" found in SQL-type databases; it is essentially a movable reference to elements
 * in an index, and it can be moved forward and backward through the list of keys.
 * <p>
 * @author  Scott Robert Ladd
 * @see     com.coyotegulch.jisp.ObjectIndex
 * @see     com.coyotegulch.jisp.IndexedObjectDatabase
 */
public class BTreeIterator implements IndexIterator
{
    private class PagePointer
    {
        public Page  m_page;
        public int   m_index;
        
        public PagePointer(Page pg, int index)
        {
            m_page  = pg;
            m_index = index;
        }
    }
    
    private boolean            m_valid = true;
    private BTreeIndex         m_index;
    private Stack              m_pageStack;
    private Page               m_currentPage;
    private int                m_currentIndex;
    
    /**
     * Creates a new <code>BTreeIterator</code> for a given index and database.
     *
     * @param   index                   the index to which this iterator is attached.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @see     com.coyotegulch.jisp.BTreeIndex
     */
    public BTreeIterator(BTreeIndex index) throws IOException, ClassNotFoundException
    {
        // save reference to the tree file
        m_index    = index;
        
        // create page stack for recursion
        m_pageStack = new Stack();
        
        // read the root page
        m_currentPage  = m_index.readRoot();
        m_currentIndex = -1;
        
        if (m_currentPage.m_header.m_numKeys < 1)
            m_valid = false;
        else
            m_valid = moveNext();
    }
    
    /**
     * Creates a new <code>BTreeIterator</code> that points to the same location as an existing
     * <code>BtreeIterator</code>.
     *
     * @param   iterator                the iterator to be copied.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @see     com.coyotegulch.jisp.BTreeIndex
     */
    public BTreeIterator(BTreeIterator iterator) throws IOException, ClassNotFoundException
    {
        // save reference to the tree file
        m_index    = iterator.m_index;
        
        // create page stack for recursion
        m_pageStack = (Stack)iterator.m_pageStack.clone();
        
        // read the root page
        m_currentPage  = m_index.readRoot();
        m_currentIndex = iterator.m_currentIndex;
    }
    
    /**
     * Returns the reference (usually a file pointer) currently associated with this iterator.
     *
     * @return  the record <code>Object</code> currently referenced by this iterator, or -1 if the iterator is invalid
     */
    public long getRecPtr() throws IOException
    {
        long result = -1;
        
        if (m_valid)
            result = m_currentPage.m_recPtr[m_currentIndex];
            
        return result;
    }
    
    /**
     * Returns the key <code>Object</code> currently associated with this iterator.
     *
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @return  the key <code>Object</code> currently referenced by this iterator, or null if the iterator is invalid
     */
    public Object getKey() throws IOException
    {
        Object result = null;
        
        if (m_valid)
        {            
            // read the object
            try
            {
                result = m_currentPage.m_key[m_currentIndex];
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
        }
        
        return result;
    }
    
    /**
     * Moves this iterator to the next key and reference in sequence. 
     *
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @return <code>true</code> if the operation was successful; <code>false</code> otherwise.
     */
    public boolean moveNext() throws IOException, ClassNotFoundException
    {
        if (!m_valid)
            return false;
            
        // next entry
        ++m_currentIndex;
        
        boolean popped = false;
        
        // have we passed the end of this page?
        if (m_currentIndex == m_currentPage.m_header.m_numKeys)
        {
            if (m_currentPage.m_link[m_currentIndex] != BTreePageFile.NULL_PTR)
            {
                // save the current page
                m_pageStack.push(new PagePointer(m_currentPage,m_currentIndex));
                
                // move to the child page
                m_currentPage  = m_index.read(m_currentPage.m_link[m_currentIndex]);
                m_currentIndex = 0;
            }
        }
        
        while (m_currentIndex >= m_currentPage.m_header.m_numKeys)
        {
            if (m_pageStack.empty())
            {
                // we're done
                return false; 
            }
            else
            {
                // go back to previous page and continue with next entry
                PagePointer temp = (PagePointer)m_pageStack.pop();
                
                m_currentPage  = temp.m_page;
                m_currentIndex = temp.m_index;
                
                popped = true;
            }
        }

        try
        {
            // go to first entry in child, if it exists
            if (!popped)
            {
                while (m_currentPage.m_link[m_currentIndex] != Page.NULL_PTR)
                {
                    // save the current page
                    m_pageStack.push(new PagePointer(m_currentPage,m_currentIndex));
                    
                    // move to the child page
                    m_currentPage  = m_index.read(m_currentPage.m_link[m_currentIndex]);
                    m_currentIndex = 0;
                }
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            return false;
        }
        
        return true;
    }
    
    /**
     * Moves this iterator to the previous key and reference in sequence. 
     *
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  BTreeException          when an error occurs during B-Tree processing
     * @return <code>true</code> if the operation was successful; <code>false</code> otherwise.
     */
    public boolean movePrevious() throws IOException, ClassNotFoundException
    {
        if (!m_valid)
            return false;
            
        boolean popped = false;
        
        // are we at the beginning of the page?
        while (m_currentIndex == 0) 
        {
            // are we done yet?
            if (m_pageStack.empty())
                return false;
                
            // go back to previous page and continue with next entry
            PagePointer temp = (PagePointer)m_pageStack.pop();
            
            m_currentPage  = temp.m_page;
            m_currentIndex = temp.m_index;
            
            popped = true;
        }
        
        // previous entry
        --m_currentIndex;
        
        if (m_currentPage.m_link[m_currentIndex] != BTreePageFile.NULL_PTR)
        {
            // save the current page
            m_pageStack.push(new PagePointer(m_currentPage,m_currentIndex));
            
            // move to the child page
            m_currentPage  = m_index.read(m_currentPage.m_link[m_currentIndex]);
            m_currentIndex = m_currentPage.m_header.m_numKeys - 1;
        }
        
        return true;
    }
    
    /**
     * Moves this iterator to the first key and reference in sequence.
     *
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  BTreeException          when an error occurs during B-Tree processing
     * @return <code>true</code> if the operation was successful; <code>false</code> otherwise.
     */
    public boolean moveFirst() throws IOException, ClassNotFoundException
    {
        // verify validity
        if (!m_valid)
            return false;
            
        // create page stack for recursion
        m_pageStack = new Stack();
        
        // read the root page
        m_currentPage  = m_index.readRoot();
        m_currentIndex = -1;
        moveNext();
        
        return true;
    }
    
    /**
     * Moves this iterator to the last key and reference in sequence.
     *
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  BTreeException          when an error occurs during B-Tree processing
     * @return <code>true</code> if the operation was successful; <code>false</code> otherwise.
     */
    public boolean moveLast() throws IOException, ClassNotFoundException
    {
        // verify validity
        if (!m_valid)
            return false;
            
        // create page stack for recursion
        m_pageStack = new Stack();
        
        // read the root page
        m_currentPage  = m_index.readRoot();
        m_currentIndex = -1;
        
        while (moveNext())
        {
            // just loop 'til we find the end
            // not efficient, but works right now when I'm testing
        }
        
        return true;
    }
    
    /**
     * Moves this iterator to point to the given <code>key</code>.
     *
     * @param   key                     key identifier to find
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  BTreeException          when an error occurs during B-Tree processing
     * @return <code>true</code> if the operation was successful; <code>false</code> otherwise.
     */
    public boolean moveTo(KeyObject key) throws IOException, ClassNotFoundException
    {
        return moveTo(key,false);
    }
    
    /**
     * Moves this iterator to point to the given <code>key</code>.
     *
     * @param   key                     key identifier to find
     * @param   acceptNext              when true, allows the search to return the next record in sequence
     *                                  if an exact match is not found
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  BTreeException          when an error occurs during B-Tree processing
     * @return <code>true</code> if the operation was successful; <code>false</code> otherwise.
     */
    public boolean moveTo(KeyObject key, boolean acceptNext) throws IOException, ClassNotFoundException
    {
        // verify validity
        if (!m_valid)
            return false;
            
        // create page stack for recursion
        m_pageStack = new Stack();
        
        // read the root page
        m_currentPage  = m_index.readRoot();
        m_currentIndex = -1;
        return search(key,m_currentPage,acceptNext);
    }
    
    private boolean search(KeyObject key, Page pg, boolean acceptNext) throws IOException, ClassNotFoundException
    {
        int pos = 0;

        if (pg.m_header.m_numKeys == 0)
            return false;
            
        int comp = KeyObject.KEY_ERROR;

        while (true)
        {
            if (pos == pg.m_header.m_numKeys)
                break;

            comp = pg.m_key[pos].compareTo(key);

            if (comp == KeyObject.KEY_EQUAL)
            {
                m_currentPage  = pg;
                m_currentIndex = pos;
                return true;
            }
            else
            {
                if (comp == KeyObject.KEY_LESS)
                    ++pos;
                else
                    break;
            }
        }

        if (pg.m_link[pos] == BTreeIndex.NULL_PTR)
        {
            if (acceptNext)
            {
                if (comp == KeyObject.KEY_MORE)
                {
                    m_currentPage  = pg;
                    m_currentIndex = pos;
                    return true;
                }
                else
                    return false;
            }
            else
                return false;
        }
        else
        {
            // save the current page
            m_pageStack.push(new PagePointer(pg,pos));
            return search(key,m_index.read(pg.m_link[pos]),acceptNext);
        }
    }
    
    /**
     * Checks to see if this iterator is valid.
     *
     * @return <code>true</code> if the iterator is valid; <code>false</code> if it is invalid.
     */
    public boolean isValid()
    {
        return m_valid;
    }
    
    /**
     * Sets this iterator's state to invalid.
     */
    public void invalidate()
    {
        m_valid = false;
    }
}
