//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  BTreePageFile.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;
import java.util.*;

class BTreePageFile extends BTreeShared
{
	// class constants
	private static final int MINIMUM_ORDER = 5;

    /**
     * Header data for a file containing B-Tree pages.
     */
    protected PageFileHeader   m_header;
    private   PageDatabaseFile m_datafile;
    
    /**
     *  A cache of pages that have already been loaded into memory. As each page is read
     */    
    protected HashMap          m_pageCache;
	
    // constructors
    BTreePageFile(String name) throws IOException, ClassNotFoundException
    {
        m_datafile = new PageDatabaseFile(name);

        // read existing m_header
        m_datafile.rewind();
        m_header = (PageFileHeader)m_datafile.readObject();
        m_pageCache = new HashMap();
    }

    BTreePageFile(String name, int order, KeyObject nullKey, boolean hasDupes) throws IOException, ClassNotFoundException
    {
        // delete any existing files
        File killthis = new File(name);
        killthis.delete();

        m_datafile = new PageDatabaseFile(name);
        m_pageCache = new HashMap();
        constructorHelper(order,nullKey,hasDupes);
    }

    private void constructorHelper(int order, KeyObject nullKey, boolean hasDupes) throws IOException, ClassNotFoundException
    {
		// adjust order to fit requirements
		if (order < MINIMUM_ORDER)
			order = MINIMUM_ORDER;
		else
		{
			// make certain order is odd
			if ((order & 1) != 1)
				++order;
		}

        // new file, new m_header
        m_header = new PageFileHeader();
		
        m_header.m_order    = order;
        m_header.m_maxKeys  = order - 1;
        m_header.m_minKeys  = order / 2;
        m_header.m_count    = 0;
        m_header.m_ticker   = 0;
        m_header.m_rootPtr  = NULL_PTR;
        m_header.m_hasDupes = hasDupes;
        m_header.m_nullKey  = nullKey;

        // write m_header to file
        m_datafile.writeObject(m_header);

        // create a new root
        Page root = new Page(m_header);

        // store root and update m_header
        root.m_header.m_filePtr = m_datafile.getNextOpen();
        m_header.m_rootPtr      = m_datafile.writeObject(root);
        
        // update m_header
        m_datafile.rewind();
        m_datafile.rewriteObject(m_header);
    }

    // methods
    void write(Page pg, boolean root) throws IOException
    {
        // if new page, find a place to write it
        if (pg.m_header.m_filePtr == NULL_PTR)
        {
            pg.m_header.m_filePtr = m_datafile.getNextOpen();

            // add page to cache
            Long cacheKey = new Long(pg.m_header.m_filePtr);
            m_pageCache.put(cacheKey,pg);
        
            // write the new page
            m_datafile.writeObject(pg);
        }
        else // write in old position
        {
            m_datafile.seek(pg.m_header.m_filePtr);
            m_datafile.rewriteObject(pg);
        }

        // if root, change root pointer
        if (root)
        {
            m_header.m_rootPtr = pg.m_header.m_filePtr;

            // update m_header
            m_datafile.rewind();
            m_datafile.rewriteObject(m_header);
        }
    }
    
    void rewriteHeader() throws IOException
    {
        m_datafile.rewind();
        m_datafile.rewriteObject(m_header);
    }

    Page read(long pos) throws IOException, ClassNotFoundException
    {
        Page result = null;
        Long cacheKey = new Long(pos);
        
        // System.err.print("--- Reading page from position " + pos);
        
        // see if page is already in cache
        if (m_pageCache.containsKey(cacheKey))
        {
            // System.err.println(" from cache");
            
            result = (Page)m_pageCache.get(cacheKey);
        }
        else
        {
            // System.err.println(" from file");
            
            // move to specified position in data file
            m_datafile.seek(pos);

            // read and return a Page
            result = (Page)m_datafile.readObject();
            
            // add new page to cache
            m_pageCache.put(cacheKey,result);
        }

        return result;
    }

    Page readRoot() throws IOException, ClassNotFoundException
    {
        return read(m_header.m_rootPtr);
    }

    void delete(Page pg) throws IOException
    {
        // remove page from cache if it is in there
        Long cacheKey = new Long(pg.m_header.m_filePtr);
        
        if (m_pageCache.containsKey(cacheKey))
            m_pageCache.remove(cacheKey);
        
        // move to specified position in data file
        m_datafile.seek(pg.m_header.m_filePtr);

        // delete this record
        m_datafile.delete();
    }
    
    HashMap getPageCache()
    {
        return m_pageCache;
    }
    
    public void close() throws IOException
    {
        if (m_datafile != null)
        {
            m_datafile.close();
            m_datafile = null;
        }
    }
}

