//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  HashIndex.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/**
 * <code>HashIndex</code> associates a <code>key</code> value with a reference to persistent data,
 * using a disk-based hash table. The index assumes that the <code>position</code> of a serializable
 * object is constant, regardless of the actual meaning of the <code>long</code> position. In
 * general, the position represents the location (file pointer) of a <code>Serializable</code>
 * object in an <code>IndexedObjectDatabase</code>.
 * <p>
 * Use the <code>IndexedObjectDatabase.attachIndex</code> method to attach a <code>BTreeIndex</code>
 * to a database. A database updates all attached indexes when it writes or removes objects
 * from its file.
 * <p>
 * <b>This class should not be used in new code.</b> It was originally design to illustrate
 * algorithms, and it suffers from a number of performance problems. A better, more efficient
 * hash-bvased index will be introduced in version 3.0 of Jisp.
 *
 * @deprecated  As of Jisp versioon 2.5.0, the HashIndex class is deprecated; it will be replaced in
 *              Jisp 3.0 by a more efficient and simpler hash-based index.
 * @author	Scott Robert Ladd
 * @see		com.coyotegulch.jisp.ObjectIndex
 */
public class HashIndex implements ObjectIndex
{
    // class storage
    private static final long NULL_POS = -1L;
    
    // working storage
    private String             m_name;
    private ObjectDatabaseFile m_file;
    private HashIndexHeader    m_header;
    
    //
    // constructors
    //
    /**
     * Creates the file as a <code>HashIndex</code>. The initial length of each list is 
     * based on <code>numBuckets</code> and <code>macRecords</code>. The null key is a "blank" key; obtain
     * a null key for any <code>KeyObject</code> type by calling the <code>KeyObject.makeNullKey</code>
     * method.
     * <p>
     * @param   name        The name of the external index file to be opened.
     * @param   numBuckets  The number of lists in the hash table; this is also known as the "hash factor."
     *                      Use prime numbers to limit the number of collisions and increase the 
     *                      efficiency of the index.
     * @param   maxRecords  The expected number of records to be indexed. This is <code>not</code> a limit
     *                      on the index size; the index will expand to handle up to 2 billion records.
     *                      However, the best performance is obtained by setting <code>maxRecords</code>
     *                      to a value that reflects the expected size of the database.
     * @param   nullKey     The null key serves two purposes: to identify a blank table entry,
     *                      and to perform type-checking on new keys added to the index. All
     *                      keys added to the index must have the same type as <code>nullkey</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @see     com.coyotegulch.jisp.KeyObject
     */    
    public HashIndex(String name, int numBuckets, int maxRecords, KeyObject nullKey) throws IOException, ClassNotFoundException
    {
        // superclass constructor
        m_name = new String(name);
        m_file = new ObjectDatabaseFile(name,true);
        
        // generate header
        m_header = new HashIndexHeader();
        
        if (numBuckets > 0)
            m_header.m_nBuckets = numBuckets;
        else
            m_header.m_nBuckets = 101;
        
        m_header.m_bucketPos = new long [m_header.m_nBuckets];
        
        for (int n = 0; n < m_header.m_nBuckets; ++n)
            m_header.m_bucketPos[n] = NULL_POS;
        
        m_header.m_nullKey = nullKey.makeNullKey();
        m_header.m_padding = maxRecords / numBuckets + 1;
        
        if (m_header.m_padding < 10)
            m_header.m_padding = 10;
        
        // write header
        m_file.rewind();
        m_file.writeObject(m_header);
    }

    /**
     * Opens a file as a <code>HashIndex</code>.
     * <p>
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @param   name    The name of the external index file to be opened.
     */    
    public HashIndex(String name) throws IOException, ClassNotFoundException
    {
        // superclass constructor
        m_file = new ObjectDatabaseFile(name,false);
        
        // read header
        m_file.rewind();
        m_header = (HashIndexHeader)m_file.readObject();
    }
    
    //
    // methods
    //
    
    /**
     * Insert a key into the database, associating a record position with the given key.
     *
     * @param   key         A key identifying the record to be read.
     * @param   pos         File position of record associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DuplicateKey            when inserting a duplicate key into an index that does not
     *                                  support duplicates
     * @see     com.coyotegulch.jisp.DuplicateKey
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void insertKey(KeyObject key, long pos) throws IOException, DuplicateKey, ClassNotFoundException
    {
        // calculate bucket
        int n = key.hashCode() % m_header.m_nBuckets;
        
        HashIndexBucket bucket = null;
        
        boolean new_bucket;
        
        // does this bucket exist?
        if (m_header.m_bucketPos[n] == NULL_POS)
        {
            // create a new bucket record
            bucket = new HashIndexBucket();
            
            bucket.m_key = new KeyObject [m_header.m_padding];
            bucket.m_pos = new long [m_header.m_padding];
            
            for (int i = 0; i < m_header.m_padding; ++i)
            {
                bucket.m_key[i] = m_header.m_nullKey.makeNullKey();
                bucket.m_pos[i] = NULL_POS;
            }
            
            bucket.m_empty = m_header.m_padding;
            
            new_bucket = true;
        }
        else
        {
            // read existing bucket
            m_file.seek(m_header.m_bucketPos[n]);
            bucket = (HashIndexBucket)m_file.readObject();
            
            new_bucket = false;
        }
        
        // do we need to extend the bucket's size?
        int where;
        
        if (bucket.m_empty == 0)
        {
            int new_len = bucket.m_key.length + m_header.m_padding;
            
            KeyObject [] new_key = new KeyObject [new_len];
            long []      new_pos = new long [new_len];
            
            for (int i = 0; i < new_len; ++i)
            {
                if (i >= bucket.m_key.length)
                {
                    new_key[i] = m_header.m_nullKey.makeNullKey();
                    new_pos[i] = NULL_POS;
                }
                else
                {
                    new_key[i] = bucket.m_key[i];
                    new_pos[i] = bucket.m_pos[i];
                }
            }
            
            // location of new entry
            where = bucket.m_key.length;

            // assign expanded arrays to bucket
            bucket.m_key   = new_key;
            bucket.m_pos   = new_pos;
            bucket.m_empty = m_header.m_padding - 1;
        
            // add new entry
            bucket.m_key[where] = key;
            bucket.m_pos[where] = pos;
            
            // delete existing bucket
            m_file.seek(m_header.m_bucketPos[n]);
            m_file.delete();
        
            // write new bucket & store position in header
            m_header.m_bucketPos[n] = m_file.writeObject(bucket);
        
            // rewrite header
            m_file.rewind();
            m_file.rewriteObject(m_header);
        }
        else
        {
            for (where = 0; bucket.m_pos[where] != NULL_POS; ++where)
                ;
        
            // add new entry
            bucket.m_key[where] = key;
            bucket.m_pos[where] = pos;
        
            --bucket.m_empty;
            
            // rewrite bucket
            if (new_bucket)
            {
                // write new bucket & store position in header
                m_header.m_bucketPos[n] = m_file.writeObject(bucket);
        
                // rewrite header
                m_file.rewind();
                m_file.rewriteObject(m_header);
            }
            else
            {
                m_file.seek(m_header.m_bucketPos[n]);
                m_file.rewriteObject(bucket);
            }
        }
    }
    
    /**
     * Replace the reference pos for the given key.
     *
     * @param   key         A key identifying the record to be read.
     * @param   pos         File position of record associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void replaceKey(KeyObject key, long pos) throws IOException, ClassNotFoundException
    {
        try
        {
            removeKey(key);
        }
        catch (KeyNotFound ex)
        {
            // ignore it
        }
        
        insertKey(key,pos);
    }
                                                           
    /**
     * If the key already exists, replace the reference pos for the given key. Otherwise, insert
     * a key into the database, associating a record position with the given key.     
     * 
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @param   key         A key identifying the record to be read.
     * @param   pos         File position of record associated with key
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void storeKey(KeyObject key, long pos) throws IOException, ClassNotFoundException
    {
        replaceKey(key,pos);
    }
    
    /**
     * Find the position of the object associated with a given key.
     * 
     * @param   key         A key identifying the record to be read.
     * @return  The position of the record associated with <code>key</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public long findKey(KeyObject key) throws KeyNotFound, IOException, ClassNotFoundException
    {
        // calculate bucket
        int n = key.hashCode() % m_header.m_nBuckets;
        
        HashIndexBucket bucket = null;
        
        // does this bucket exist?
        if (m_header.m_bucketPos[n] == NULL_POS)
            throw new KeyNotFound();
        
        // read bucket
        m_file.seek(m_header.m_bucketPos[n]);
        bucket = (HashIndexBucket)m_file.readObject();
        
        // search for key
        boolean found = false;
        int i;
        
        for (i = 0; i < bucket.m_key.length; ++i)
        {
            if ((bucket.m_key[i] != null) && (key.compareTo(bucket.m_key[i]) == KeyObject.KEY_EQUAL))
            {
                found = true;
                break;
            }
        }
        
        if (!found)
            throw new KeyNotFound();
        
        // return the associated record position
        return bucket.m_pos[i];
    }
    
    /**
     * Removes the given key from the index.
     *
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @param   key         A key identifying the record to be read.
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void removeKey(KeyObject key) throws IOException, ClassNotFoundException
    {
        // calculate bucket
        int n = key.hashCode() % m_header.m_nBuckets;
        
        HashIndexBucket bucket = null;
        
        // does this bucket exist?
        if (m_header.m_bucketPos[n] == NULL_POS)
            throw new KeyNotFound();
        
        // read bucket
        m_file.seek(m_header.m_bucketPos[n]);
        bucket = (HashIndexBucket)m_file.readObject();
        
        // search for key
        boolean found = false;
        int i;
        
        for (i = 0; i < bucket.m_key.length; ++i)
        {
            if ((bucket.m_key[i] != null) && (key.compareTo(bucket.m_key[i]) == KeyObject.KEY_EQUAL))
            {
                found = true;
                break;
            }
        }
        
        if (!found)
            throw new KeyNotFound();
        
        // remove key
        bucket.m_key[i] = m_header.m_nullKey.makeNullKey();
        bucket.m_pos[i] = NULL_POS;
        ++bucket.m_empty;
        
        // delete existing bucket
        m_file.seek(m_header.m_bucketPos[n]);
        m_file.delete();
        
        // write new bucket & store position in header
        m_header.m_bucketPos[n] = m_file.writeObject(bucket);
        
        // rewrite header
        m_file.rewind();
        m_file.rewriteObject(m_header);
    }
    
    /**
     * Improves the performance of this <code>HashIndex</code> by optimizing the
     * index contents.
     *
     */
    public void optimize() throws IOException, ClassNotFoundException
    {
        // create new file
        String temp_name = "" + System.currentTimeMillis() + ".tmp";
        ObjectDatabaseFile new_file = new ObjectDatabaseFile(temp_name,true);
        
        // create a new header
        HashIndexHeader new_header = new HashIndexHeader();

        new_header.m_nBuckets  = m_header.m_nBuckets;
        new_header.m_nullKey   = m_header.m_nullKey.makeNullKey();
        new_header.m_padding   = m_header.m_padding;
        new_header.m_bucketPos = (long [])m_header.m_bucketPos.clone();
        
        // rewrite header
        new_file.rewind();
        new_file.writeObject(new_header);
        
        // copy records
        for (int n = 0; n < m_header.m_nBuckets; ++n)
        {
            // read old record
            if (m_header.m_bucketPos[n] != NULL_POS)
            {
                m_file.seek(m_header.m_bucketPos[n]);
                HashIndexBucket bucket = (HashIndexBucket)m_file.readObject();
            
                // update header
                new_header.m_bucketPos[n] = new_file.writeObject(bucket);
            }
        }
        
        // rewrite header
        new_file.rewind();
        new_file.rewriteObject(new_header);
        
        // close new file
        new_file.close();
        
        // delete old file
        m_file.close();
        m_file = null;
        
        File f_old = new File(m_name);
        f_old.delete();
        
        // rename new file
        File f_new = new File(temp_name);
        f_new.renameTo(f_old);
        
        // reopen new file as current file
        m_file = new ObjectDatabaseFile(m_name,false);
        
        // use new header
        m_header = new_header;
    }
    
    public void close()  throws IOException
    {
        if (m_file != null)
        {
            m_file.close();
            m_file = null;
        }
    }
}
