//-----------------------------------------------------------------------
//  Java Indexed Serialization Package
//  ----------------------------------
//  Huffman.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/**
 * Compresses and decompresses objects using the Huffman algorithm. Huffman
 * encoding creates a set of codes for which the shortest code represents the
 * most common piece of data. Codes created by the Huffman algorithm require
 * a file to be analyzed, counting bytes to determine their frequency. From
 * the frequencies, the Huffman algorithm builds a table of codes used to
 * compress the information. Including the table of codes with the compressed
 * data allows the original file to be reconstructed.
 * <p>
 * To be most effective, Huffman encoding uses a variable length code, where
 * no code is a prefix of another, which makes decompression easier by allowing 
 * the extraction of the file bit-by-bit. The shortest codes are assigned to 
 * the most common characters, with infrequent characters receiving longer codes.
 * <p>
 * Objects to be compressed must be Serializable.
 *
 * @author  Scott Robert Ladd
 * @see     com.coyotegulch.jisp.HuffmanEncoded
 */
public class Huffman 
{
    /**
     * Encodes an object using Huffman compression.
     * 
     * @param   obj                     serializable object to be compressed
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @return                          a new <code>HuffmanEncoded</code> object containing the compressed
     *                                  object and its encoding table
     * @see     HuffmanEncoded
     */
    public static HuffmanEncoded encode(Object obj) throws IOException
    {
        // serialize object
        if ((obj instanceof Serializable) || (obj instanceof Externalizable))
        {
            // create an output byte array
            ByteArrayOutputStream raw = new ByteArrayOutputStream();

            // create an object Output stream
            ObjectOutputStream out = new ObjectOutputStream(raw);

            // write object to output stream
            out.writeObject(obj);
            out.flush();

            // encode bytes
            return compress(raw.toByteArray());
        }
        else
            throw new HuffmanException("can't encode non-serializable object");
    }

    /**
     * Decodes a compressed object using the Huffman algorithm.
     * 
     * @param   enc                     object to be decompressed
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  when an explicit cast fails
     * @return                          the decompressed <code>Object</code>
     * @see     HuffmanEncoded
     */
    public static Object decode(HuffmanEncoded enc) throws IOException, ClassNotFoundException
    {
        // decode compressed data
        byte [] decoded = decompress(enc);

        // create byte input stream from decoded bytes
        ByteArrayInputStream raw = new ByteArrayInputStream(decoded);

        // create an object input stream
        ObjectInputStream in = new ObjectInputStream(raw);

        // return the object
        return in.readObject();
    }

    private static void heapAdjust(int [] freq, int [] heap, int n, int k)
    {
        // this function compares the values in the array
        // 'freq' to order the elements of 'heap' according
        // in an inverse heap.
        int j;

        int v = heap[k - 1];

        while (k <= (n / 2))
        {
            j = k + k;

            if ((j < n) && (freq[heap[j - 1]] > freq[heap[j]]))
                ++j;

            if (freq[v] < freq[heap[j - 1]])
                break;

            heap[k - 1] = heap[j - 1];
            k = j;
        }

        heap[k - 1] = v;
        }

    // Huffman compression function
    private static HuffmanEncoded compress(byte [] info)
    {
        // get length of info
        int bytes = info.length;

        // allocate data space
        byte [] edata = new byte [bytes + 1];

        // allocate frequency table
        int [] freq = new int [512];

        // allocate heap
        int [] heap = new int [256];

        // allocate link array
        int [] link = new int [512];

        // create work area
        HuffmanEncoded encoded = new HuffmanEncoded();

        int   [] code = encoded.m_header.m_code;
        short [] clen = encoded.m_header.m_codeLength;

        // count frequencies
        int i;

        byte a = -1;

        for (i = 0; i < bytes; ++i)
            ++freq[info[i] < 0 ? 256 + info[i] : info[i]];

        // create indirect heap based on frequencies
        int n = 0; 

        for (i = 0; i < 256; ++i)
        {
            if (freq[i] != 0)
            {
                heap[n] = i;
                ++n;
            }
        }

        for (i = n; i > 0; --i)
            heapAdjust(freq,heap,n,i);

        // generate a trie from heap
        int temp;

        // at this point, n contains the number of characters
        // that occur in the info array
        while (n > 1)
        {
            // take first item from top of heap
            --n;
            temp    = heap[0];
            heap[0] = heap[n];

            // adjust the heap to maintain properties
            heapAdjust(freq,heap,n,1);

            // in upper half of freq array, store sums of 
            // the two smallest frequencies from the heap
            freq[256 + n] = freq[heap[0]] + freq[temp];
            link[temp]        =  256 + n; // parent
            link[heap[0]]     = -256 - n; // left child
            heap[0]           =  256 + n; // right child

            // adjust the heap again
            heapAdjust(freq,heap,n,1);
        }

        link[256 + n] = 0;

        // generate codes
        int j, k, x, maxx = 0, maxi = 0;
        int l;

        for (k = 0; k < 256; ++k)
        {
            if (freq[k] == 0) // character does not occur
            {
                code[k] = 0;
                clen[k] = 0;
            }
            else
            {
                i = 0;       // length of current code
                j = 1;       // bit being set in code
                x = 0;       // code being built
                l = link[k]; // link in trie
    
                while (l != 0) // while not at end of trie
                {
                    if (l < 0) // left link (negative)
                    {
                        x +=  j; // insert 1 into code
                        l  = -l; // reverse sign
                    }

                    l  = link[l]; // move to next link
                    j <<= 1;      // next bit to be set
                    ++i;          // increment code length
                }

                code[k] = x; // save code
                clen[k] = (short)i; // save code len

                // keep track of biggest key
                if (x > maxx)
                    maxx = x;

                // keep track of longest key
                if (i > maxi)
                    maxi = i;
            }
        }

        // make sure longest codes fit in unsigned long-bits
        if (maxi > 32)
            throw new HuffmanException("code size overflow");

        // encode data
        int  mask;        // mask for extracting bits
        int  nout =  0;   // number of bytes output
        byte bout =  0;   // byte of encoded data
        int  bit  = -1;   // count of bits stored in bout

        // watch for one-value file!
        if (maxx == 0)
        {
            edata[0] = (byte)info[0];
            nout = 1;
            encoded.m_header.m_oneValue = true;
        }
        else
        {
            for (j = 0; j < bytes; ++j)
            {
                int ibyte = info[j] < 0 ? 256 + info[j] : info[j];

                // start copying at first bit of code
                mask = 1 << (clen[ibyte] - 1);

                // copy code bits
                for (i = 0; i < clen[ibyte]; ++i)
                {
                    if (bit == 7)
                    {
                        // store full output byte
                        edata[nout] = bout;
                        ++nout;

                        // check for overflow
                        if (nout == bytes)
                            throw new HuffmanException("compression overflow");

                        bit  = 0;
                        bout = 0;
                    }
                    else
                    {
                        // move to next bit
                        ++bit;
                        bout <<= 1;
                    }

                    if ((code[ibyte] & mask) != 0)
                        bout |= 1;

                    mask >>>= 1;
                }
            }

            // output any incomplete bytes
            bout <<= (7 - bit);
            edata[nout] = bout;
            ++nout;
        }

        // store all information
        encoded.m_header.m_originalLength = bytes;
        encoded.m_header.m_dataLength = nout;

        byte [] edata2 = new byte [nout];
        
        for (i = 0; i < nout; ++i)
            edata2[i] = edata[i];
        
        encoded.m_compressedData = edata2;

        // return result
        return encoded;
    }

    // decompress encoded data
    private static byte [] decompress(HuffmanEncoded data)
    {
        int  i, j, n, mask, k, t;
        byte c;

        // allocate output array
        byte [] dec = new byte [data.m_header.m_originalLength];

        // handle decompression of one-value file
        if (data.m_header.m_oneValue)
        {
            for (i = 0; i < dec.length; ++i)
                dec[i] = data.m_compressedData[0];

            return dec;
        }

        // allocate heap
        int [] heap = new int [256];

        // allocate output character buffer
        byte [] outc = new byte [256];

        // initialize work areas
        int   [] code = data.m_header.m_code;
        short [] clen = data.m_header.m_codeLength;

        // create decode table as trie heap
        for (j = 0; j < 256; ++j)
        {
            outc[j] = (byte)j;
    
            // if code exists for this byte
            if ((code[j] | clen[j]) != 0)
            {
                // begin at first code bit
                k = 0;
                mask = 1 << (clen[j] - 1);

                // find proper node, using bits in code as path.
                for (i = 0; i < clen[j]; ++i)
                {
                    k = k * 2 + 1; // right link
    
                    if ((code[j] & mask) != 0)
                        ++k; // go left

                    mask >>>= 1; // next bit
                }

                heap[j] = k; // store link in heap
            }
        }

        // sort outc based on heap
        for (i = 1; i < 256; ++i)
        {
            t = heap[i];
            c = outc[i];
            j = i;

            while ((j != 0) && (heap[j - 1] > t))
            {
                heap[j] = heap[j - 1];
                outc[j] = outc[j - 1];
                --j;
            }

            heap[j] = t;
            outc[j] = c;
        }

        // find first character in table
        for (j = 0; heap[j] == 0; ++j) ;

        // decode data
        k    = 0; // link in trie
        i    = j; 
        mask = 0x80;
        n    = 0;
        int cptr = 0;
        int dptr = 0;

        while (n < data.m_header.m_originalLength)
        {
            k = k * 2 + 1; // right link

            if ((data.m_compressedData[cptr] & mask) != 0)
                ++k; // left link if bit on

            // search heap until link >= k
            while (heap[i] < k)
                ++i;

            // code matches, character found
            if (k == heap[i])
            {
                dec[dptr] = outc[i];
                ++dptr;
                ++n;
                k = 0;
                i = j;
            }

            // move to next bit
            if (mask > 1)
                mask >>>= 1;
            else
            { // code extends into next byte
                mask = 0x80;
                ++cptr;
            }
        }

        // return decoded data
        return dec;
    }
}

