//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  IndexedObjectDatabase.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.util.Vector;
import java.io.*;

/**
 * The <code>IndexedObjectDatabase</code> class provides a mechanism for
 * using a key value to an object serialized to a file.
 * <p>
 * A <code>IndexedObjectDatabase</code> object encapsulates access to a
 * <code>ObjectDatabaseFile</code> via a set of <code>ObjectIndex</code>s.
 * When an object is stored, it is associated with a key value by storing the
 * key and an object reference in each attached <code>ObjectIndex</code>.
 * Objects may then be retrieved either by providing a key and index pair,
 * or through an iterator created for an index associated with the database.
 * <p>
 * <code>Rationale</code>: The Java Database Connection (JDBC) is
 * often overkill for many applications. Sun designed JDBC to access
 * enterprise-level database systems, such as DB2, SQL Server, and Oracle.
 * Those systems, while very flexible and expansive, come with a high
 * price tag, both in terms of system requirements and database overhead.
 * <code>BTreeDatabase</code> provides a simpler tool for associating
 * "key" information with data in external storage.
 * <p>
 * @author	Scott Robert Ladd
 * @see		com.coyotegulch.jisp.BTreeIndex
 * @see		com.coyotegulch.jisp.DatabaseException
 * @see		com.coyotegulch.jisp.KeyObject
 * @see		com.coyotegulch.jisp.ObjectIndex
 */
public class IndexedObjectDatabase extends ObjectDatabaseFile
{
    // working storage
    private Vector m_indexes = new Vector();
    private IndexedObjectDatabase m_this = null;
    
    // exceptions
    private static final DatabaseException err_no_indexes    = new DatabaseException("no indexes associated with this database");
    private static final DatabaseException err_invalid_index = new DatabaseException("given index is not associated with this database");

    /**
     * Opens an existing<code> IndexedObjectDatabase</code>, using a supplied file name.
     *
     * @param   filename                The name of an external database file containing serialized objects.
     *                                  If <code>filename</code> exists, it is opened; otherwise, the constructor
     *                                  creates the file and initializes it.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    public IndexedObjectDatabase(String filename, boolean is_new) throws IOException
    {
        super(filename, is_new);
        m_this = this;
    }
    
    /**
     * Attaches an index object to this database. This index will be updated for every
     * object recorded in the database.
     *
     * @param   index       An <code>ObjectIndex</code> to be associated with this database.
     * @see     com.coyotegulch.jisp.ObjectIndex
     */
    public void attachIndex(ObjectIndex index)
    {
        m_indexes.addElement(index);
    }
    
    /**
     * Removes the association of an index with this database. The index will no longer be updated for
     * objects recorded in the database.
     *
     * @param   index       An <code>ObjectIndex</code> that should no longer be associated with this database.
     * @see     com.coyotegulch.jisp.ObjectIndex
     */
    public void removeIndex(ObjectIndex index)
    {
        m_indexes.removeElement((Object)index);
    }
    
    /**
     * Write a new object to the database, associating it with the provided keys.
     *
     * @param   key                     The key values associated with <code>obj</code>. There must be
     *                                  one key for each attached index; the keys are associated with
     *                                  the indexes in order of attachment; in other words, the first key
     *                                  is associated with the first index attached, etc.
     * @param   obj                     The object to be stored in the database.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DatabaseException       when an error occurs during database processing
     * @see		com.coyotegulch.jisp.DatabaseException
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void insert(KeyObject key[], Serializable obj) throws IOException, ClassNotFoundException
    {
        // validate this operation
        validateOperation(null);

        // write the data record
        long pos = writeObject(obj);

        // insert key into tree
        for (int n = 0; n < m_indexes.size(); ++n)
        {
            ObjectIndex index = (ObjectIndex)m_indexes.elementAt(n);
            index.insertKey(key[n],pos);
        }
    }
    
    /**
     * Writes an object to the database, associating it with the provided key, replacing
     * an existing object with a new one.
     *
     * @param   key                     key values associated with <code>obj</code>. There must be
     *                                  one key for each attached index; the keys are associated with
     *                                  the indexes in order of attachment; in other words, the first key
     *                                  is associated with the first index attached, etc.
     * @param   obj                     record object to be stored in the database.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DatabaseException       when an error occurs during database processing
     * @see	    com.coyotegulch.jisp.DatabaseException
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void write(KeyObject key[], Serializable obj) throws IOException, ClassNotFoundException
    {
        // validate this operation
        validateOperation(null);

        try
        {
            // get first index
            ObjectIndex index = (ObjectIndex)m_indexes.elementAt(0);
            
            // find the object
            long delpos = index.findKey(key[0]);
            
            // delete old record
            seek(delpos);
            delete();
        }
        catch (KeyNotFound knf)
        {
            // ignore this exception
        }

        // write the data record
        long pos = writeObject(obj);
        
        // insert key into tree
        for (int n = 0; n < m_indexes.size(); ++n)
        {
            ObjectIndex index = (ObjectIndex)m_indexes.elementAt(n);
            index.insertKey(key[n],pos);
        }
    }
    
    /**
     * Read the object associated with a given key.
     *
     * @param   key                     key identifying the record to be read
     * @param   index                   index used to retrieve the object by <code>key</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DatabaseException       when an error occurs during database processing
     * @return                          The record object associated with <code>key</code>, or <code>null</code>
     *                                  if no such object could be found.
     * @see		com.coyotegulch.jisp.DatabaseException
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public Object read(KeyObject key, ObjectIndex index) throws IOException, ClassNotFoundException
    {
        // validate this operation
        validateOperation(index);
        
        long pos;

        try
        {
            pos = index.findKey(key);
        }
        catch (KeyNotFound knf)
        {
            // return the object if the key was not found
            return null;
        }
        
        // read the object
        try
        {
            seek(pos);
            return readObject();
        }
        catch (Exception ex)
        {
            System.err.println("key = " + key + ", pos = " + pos);
            ex.printStackTrace();
            return null;
        }
    }
    
    /**
     * Read the object associated with a given iterator.
     *
     * @param   iterator                a <code>BTreeIterator</code> pointing to the requested record
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DatabaseException       when an error occurs during database processing
     * @return                          The record object associated with <code>iterator</code>, or <code>null</code>
     *                                  if the iterator was invalid or no such record could be found.
     * @see		com.coyotegulch.jisp.DatabaseException
     * @see     com.coyotegulch.jisp.BTreeIterator
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public Object read(IndexIterator iterator) throws IOException, ClassNotFoundException
    {
        // validate this operation
        if (iterator.isValid())
        {
            // read the object
            try
            {
                seek(iterator.getRecPtr());
                return readObject();
            }
            catch (Exception ex)
            {
                System.err.println("key = " + iterator.getKey() + ", recptr = " + iterator.getRecPtr());
                ex.printStackTrace();
            }
        }
            
        return null;
    }
    
    /**
     * Delete the record associated with a given key.
     * <p>
     * @param   key                     array of keys identifying the record to be deleted.
     * @param   index                   index used to retrive the object by <code>key</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DatabaseException       when an error occurs during database processing
     * @see		com.coyotegulch.jisp.DatabaseException
     * @see     com.coyotegulch.jisp.KeyObject
     */
    public void remove(KeyObject key[]) throws IOException, DatabaseException, ClassNotFoundException
    {
        // validate this operation
        validateOperation(null);
        
        // find record to be deleted
        ObjectIndex index = (ObjectIndex)m_indexes.elementAt(0);
        
        long pos = index.findKey(key[0]);

        // delete data record
        seek(pos);
        delete();
        
        // must remove key/record from ALL associated indexes...
        for (int n = 0; n < m_indexes.size(); ++n)
        {
            index = (ObjectIndex)m_indexes.elementAt(n);
            index.removeKey(key[n]);
        }
    }

    // validate a given operation    
    private void validateOperation(ObjectIndex index) throws DatabaseException, IOException, ClassNotFoundException
    {
        // check for indexes
        if (m_indexes.size() <= 0)
            throw err_no_indexes;
        
        // make sure this index is associated with this file
        if ((index != null) && (!m_indexes.contains(index)))
            throw err_invalid_index;
    }
    
}
