//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  KeyObject.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/** * <code>KeyObject</code> defines an characteristics of an object to be used as a key * in an <code>ObjectIndex</code>. Since Java lacks true generic types, it is up to * the programmer to ensure that all <code>KeyObject</code>s have the same concrete type * for a given <code>ObjectIndex</code>. Each <code>ObjectIndex</code> supports a single * concrete <code>KeyObject</code> type. * <p> * Concrete <code>KeyObject</code> classes must always produce a  
 * <p>
 * @author	Scott Robert Ladd
 * @see		com.coyotegulch.jisp.ObjectIndex
 */
abstract public class KeyObject implements Externalizable
{
    static final long serialVersionUID = -1183464016236393879L;
    
    //
    // comparison constants
    //
    
    /**
     * Returned by <code>compareTo</code> when the invoking key is less than the parameter key.
     * <p>
     */
    public static final int KEY_LESS  = -1;
    
    /**
     * Returned by <code>compareTo</code> when the invoking key is greater than the parameter key.
     * <p>
     */
    public static final int KEY_EQUAL =  0;
    
    /**
     * Returned by <code>compareTo</code> when the invoking key equals the parameter key.
     * <p>
     */
    public static final int KEY_MORE  =  1;
    
    /**
     * Returned by <code>compareTo</code> when the invoking key has a different type from the parameter key.
     * <p>
     */
    public static final int KEY_ERROR = 99;

    //
    // methods
    //
    
    /**
     * Compares the invoking object to <code>key</code>, returning one of the <code>KEY_*</code>
     * constants based on the relationship of the two keys.
     * <p>
     * @param   key The key value to be compared against the invoking key.     * @return  One of the <code>KEY_*</code> constants based on the relationship of the two keys.     */
    public abstract int compareTo(KeyObject key);
    
    /**
     * Returns a "null" (i.e., blank) key, to be used as a space-holder in certain kinds of
     * <code>ObjectIndexes</code>.
     * <p>
     * For example, a <code>BTreeIndex</code> maintains a set of index pages that must contain
     * a set number of fixed-length keys. When creating a <code>BTreeIndex</code>, you must provide
     * a "null" key, using the <code>makeNullKey</code> method. 
     * <p>
     * @return  A "null" (i.e., blank) key constant.
     * @see     com.coyotegulch.jisp.BTreeIndex     */
    public abstract KeyObject makeNullKey();
	
    /**
     * Returns the <code>String</code> representation of a <code>KeyObject</code>.
     * <p>
     * @return  The <code>String</code> representation of a <code>KeyObject</code>.
     */
	public String toString()
	{
		return "KeyObject";	
	}
}
