//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  LongKey.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/** * <code>LongKey</code> defines a key based on the <code>long</code> type.
 * 
 * @author	Scott Robert Ladd
 * @see		com.coyotegulch.jisp.KeyObject
 * @see		com.coyotegulch.jisp.ObjectIndex
 */
public class LongKey extends KeyObject
{
    static final long serialVersionUID = 3500210952961277131L;
        
    //
    // working storage
    //
    private long m_key;
    
    /**
     * Creates a new <code>LongKey</code> with a specific value.
     *
     * @param   keyValue   value of the newly-created key.
     */    
    public LongKey(long key_value)
    {
        m_key = key_value;
    }
    
    /**
     * Creates a blank <code>LongKey</code> (value = 0).
     */    
    public LongKey()
    {
        m_key = 0;
    }
    
    /**
     * Compares the invoking <code>LongKey</code> to <code>key</code>, returning one of the
     * <code>KEY_*</code> constants based on the relationship of the two keys.
     * <p>
     * @param   key The <code>LongKey</code> value to be compared against the invoking key.     
     * @return  One of the <code>KEY_*</code> constants based on the relationship of the two keys.     
     */
    public int compareTo(KeyObject key)
    {
        if (key instanceof LongKey)
        {
            
            long comp = m_key - ((LongKey)key).m_key;
    
            if (comp == 0)
                return KEY_EQUAL;
            else
                if (comp < 0)
                    return KEY_LESS;
                else
                    return KEY_MORE;
        }
        else
            return KEY_ERROR;
    }
    
    /**
     * Returns a "null" (i.e., blank) <code>LongKey</code>, to be used as a space-holder in certain kinds of
     * <code>ObjectIndexes</code>. The "null" value is 0xFFFFFFFFFFFFFFFF (all ones).
     *
     * @return  A "null" (i.e., blank) <code>LongKey</code> constant.
     * @see     com.coyotegulch.jisp.BTreeIndex     */
    public KeyObject makeNullKey()
    {
        LongKey nullKey = new LongKey();
        nullKey.m_key = 0xFFFFFFFFFFFFFFFFL;
        return nullKey;
    }
    
    /**
     * <code>LongKey</code> implements the <code>writeExternal</code> method to save its
     * contents.
     */
    public void writeExternal(ObjectOutput out) throws IOException
    {
        out.writeObject(new Long(m_key));
    }
    
    /**
     * <code>LongKey</code> implements the readExternal method to restore its contents.
     */
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException
    {
        m_key = ((Long)(in.readObject())).longValue();
    }

    /**
     * Returns the <code>long</code> value of a <code>LongKey</code>.
     *
     * @return  The <code>long</code> value of a <code>LongKey</code>.
     */
    public long longValue()
    {
        return m_key;
    }

    /**
     * Returns the <code>String</code> representation of a <code>LongKey</code>.
     *
     * @return  The <code>String</code> representation of a <code>LongKey</code>.
     */
    public String toString()
    {
        return Long.toString(m_key);
    }
    
    /**
     * Returns a hash code value for the object, which is, in effect, the     
     * the value of the key.
     *
     * @return  a hash code value for this object.
     * @see     java.lang.Object     
     * @see     com.coyotegulch.jisp.HashIndex
     */
    public int hashCode()
    {
        return (int)m_key;
    }
}
