//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  ObjectDatabaseFile.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/**
 * The <code>ObjectDatabaseFile</code> provides a random-access file that 
 * serializes objects to variable length records in a random-access file.
 * An <code>ObjectDatabaseFile</code> allows the deletion of records and
 * automatically uses such space for new records.
 * <p>
 * For the purposes of an <code>ObjectDatabaseFile</code>, consider <i>record</i>
 * and <i>object</i> to be synonyms. In traditional database terms, an
 * <code>ObjectDatabaseFile</code> is a <i>table</i>, each object is a <i>row</i>,
 * and each field in an object is a <i>column</i>. Unlike a full-featured
 * database, an <code>ObjectDatabaseFile</code> can contain a set of heterogenous
 * records -- in other words, different objects can be stored together in one
 * <code>ObjectDatabaseFile</code>. This class is the foundation of the Jisp
 * package, providing the fundamental file structures for both indexes and data 
 * files.
 * <p>
 * You can store any serializable object in an <code>ObjectDatabaseFile</code>.
 * Records in an <code>ObjectDatabaseFile</code> can have varying lengths based 
 * on the serialization requirements of objects. To handle variable-length records, 
 * a <code>ObjectDatabaseFile</code> records the size of each record. When a record
 * is deleted, the space it occupied is marked as empty; the files maintains a 
 * linked list of deleted record locations and thier sizes.
 * <p>
 * Inserting a new record involves a traverse of the deleted list, looking for an
 * empty record that is large enough to contain the new information. If the deleted 
 * list is empty, or the new record is too large to fit into any open slots, the 
 * new object record is appended to the file.
 * <p>
 * Reusing deleted record space has a drawback: it leaves dead space in the file when
 * a newly-inserted record is smaller than the &quot;deleted&quot; space it 
 * overwrites. Deleted records also use space in the file until a new record is 
 * written into their location. Periodically, it makes sense to <i>compact</i> the 
 * file, removing the wasted space and eliminating deleted records. <code>ObjectDatabaseFile</code> 
 * recovers "waste space" by re-writing the file, record by record, using exact record lengths, 
 * ignoring deleted records, and calling a user-provided method so that indexes can be
 * simultaneously regenerated.
 *
 * @author  Scott Robert Ladd
 * @see com.coyotegulch.jisp.IndexedObjectDatabase
 * @see com.coyotegulch.jisp.PageDatabaseFile
 */
public class ObjectDatabaseFile
{

    /**
     * Marks the end of the list of deleted records
     */
    protected static final long DEL_LIST_END = -1;

    /**
     * Marks a given record as deleted
     */
    protected static final byte IS_DELETED = 0;

    /**
     * Marks a given record as containing live data
     */
    protected static final byte IS_ACTIVE  = 1;

    /**
     * Physical file containing serialized objects
     */
    protected RandomAccessFile m_dataFile;

    /**
     * Name of the underlying physical file
     */
    protected String           m_fileName;

    /**
     * File position of the first delete record
     */
    protected long             m_firstDeleted;

    /**
     * Filter to be applied to each object read or written
     */
    protected RecordFilter     m_filter;

    /**
     * Creates or opens a file that stores objects in variable-length records. 
     * 
     * @param   name                    name of the file to be created
     * @param   is_new                  <code>true</code> when a new file should be created, overwriting
     *                                  any existing file with the same name; <code>false</code> when
     *                                  opening and using an existing file
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    public ObjectDatabaseFile(String name, boolean is_new) throws IOException
    {
        // delete any existing files
        if (is_new)
        {
            File killthis = new File(name);
            killthis.delete();
        }
        
        m_fileName = new String(name);
        m_dataFile = new RandomAccessFile(name,"rw");
        m_filter   = null;
        constructorHelper();
    }

    /**
     * Creates or opens a file that stores objects in variable-length records. 
     * 
     * @param   file                    indentity of the file to be created
     * @param   is_new                  <code>true</code> when a new file should be created, overwriting
     *                                  any existing file with the same identity; <code>false</code> when
     *                                  opening and using an existing file
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    public ObjectDatabaseFile(File file, boolean is_new) throws IOException
    {
        // delete any existing files
        if (is_new)
            file.delete();
        
        m_fileName = new String(file.getName());
        m_dataFile = new RandomAccessFile(file,"rw");
        constructorHelper();
    }

    private void constructorHelper() throws IOException
    {
        // check for new file
        if (m_dataFile.length() == 0)
        {
            // write a new file header
            m_dataFile.writeLong(DEL_LIST_END);
            m_firstDeleted = DEL_LIST_END;            
        }
        else
            m_firstDeleted = m_dataFile.readLong();
    }

    /**
     * Writes a serializable object to the file. <code>writeObject</code> stores the
     * new record in the first &quot;deleted&quot; location of sufficient size, or
     * it appends the new record if suitable deleted space is not available.
     * 
     * @param   obj                     serializable object to be written as a record
     * @return                          file position of the newly-written record
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    // write object to current position
    public long writeObject(Serializable obj) throws IOException
    {
        // create a byte stream to hold the encoded object
        SerialObjectOutputStream bytes = new SerialObjectOutputStream();

        // create a stream to write the object
        ObjectOutputStream ostrm = createObjectOutputStream(bytes);

        // write the object
        if (m_filter == null)
            ostrm.writeObject(obj);
        else
            ostrm.writeObject(m_filter.filterEncode(obj));

        // ensure that the entire object is written
        ostrm.flush();

        // get the length of the output data
        int  datalen = bytes.getSize(); // data length
        int  reclen  = datalen;      // rec length
        long pos;                    // beginning of rec header

        // find someplace to put the record
        if (m_firstDeleted == DEL_LIST_END) // append
        {
            // find end of file
            pos = m_dataFile.length();

            // move to end of the file
            m_dataFile.seek(pos);
        }
        else // replace a deleted record
        {
            // start with first deleted record
            long prev = DEL_LIST_END;

            pos = m_firstDeleted;

            while (true)
            {
                // move to the record
                m_dataFile.seek(pos);

                // read header info
                byte deleted = m_dataFile.readByte();
				m_dataFile.readInt();
                reclen       = m_dataFile.readInt();
                long next    = m_dataFile.readLong();

                // check integrity
                if (deleted == IS_ACTIVE)
                    throw new IOException("corrupt delete list");

                // is deleted record big enough for new data?
                if (reclen >= datalen)
                {
                    // adjust chain by changing next to prev
                    if (prev == DEL_LIST_END)
                    {
                        // change file header deletion mark
                        m_firstDeleted = next;

                        // write new deleted record # at beginning of file
                        m_dataFile.seek(0);
                        m_dataFile.writeLong(m_firstDeleted);
                    }
                    else // go back to previous record
                    {
                        // read and change prev header
                        m_dataFile.seek(prev + 1 + 4 + 4);
                        m_dataFile.writeLong(next);
                    }

                    m_dataFile.seek(pos);

                    break;
                }

                // deleted record is too small; go to next (or end)
                prev = pos;

                // do we have any more deleted records in the chain?
                if (next == DEL_LIST_END)
                {
                    // append to end of file
                    pos = m_dataFile.length();
                    m_dataFile.seek(pos);
                    break;
                }
                else
                    pos = next;
            }
        }

        // make sure this record is marked as not deleted
        m_dataFile.writeByte(IS_ACTIVE);

        // write the length of the output
        m_dataFile.writeInt(datalen);

        // write the length of the record
        m_dataFile.writeInt(reclen);

        // write the "deleted" pointer
        m_dataFile.writeLong(DEL_LIST_END);

        // store the object to the actual file
       bytes.writeTo(m_dataFile);

        // return length of output
        return pos;
    }

    /**
     * Re-writes a serializable object to the file, at the current file position.
     * 
     * @param   obj                     serializable object to be written as a record
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    // write an object at the current position
    public void rewriteObject(Serializable obj) throws IOException
    {
        // create a byte stream to hold the encoded object
        SerialObjectOutputStream bytes = new SerialObjectOutputStream();
        
        // create a stream to write the object
        ObjectOutputStream ostrm = createObjectOutputStream(bytes);

        // write the object
        if (m_filter == null)
            ostrm.writeObject(obj);
        else
            ostrm.writeObject(m_filter.filterEncode(obj));

        // ensure that the entire object is written
        ostrm.flush();

        // get the length of the output data
        int  datalen = bytes.getSize(); // data length
        int  reclen  = datalen;      // rec length

        // make sure this record is marked as not deleted
        m_dataFile.writeByte(IS_ACTIVE);

        // write the length of the output
        m_dataFile.writeInt(datalen);

        // write the length of the record
        m_dataFile.writeInt(reclen);

        // write the "deleted" pointer
        m_dataFile.writeLong(DEL_LIST_END);

        // store the object to the actual file
       bytes.writeTo(m_dataFile);
    }

    /**
     * 
     */
    // read object from current record
    public Object readObject() throws IOException, ClassNotFoundException
    {
        while (true)
        {
            // check record status
            byte deleted = m_dataFile.readByte();

            // break from loop if this is an active record
            if (deleted == IS_ACTIVE)
                break;

            // skip length of object
            m_dataFile.skipBytes(4);

            // get the length of the object
            int n = m_dataFile.readInt();

            // seek to beginning of next record
            m_dataFile.skipBytes(n + 8);
        }

        // get the length of the object
        int datalen = m_dataFile.readInt();

        // get length of record
        int reclen = m_dataFile.readInt();

        // skip the "deleted" pointer
        m_dataFile.skipBytes(8);

        // retrieve the object data and store it
        byte [] data = new byte [datalen];
        m_dataFile.readFully(data);

        // position pointer to beginning of next record
        int diff = reclen - datalen;

        if (diff > 0)
            m_dataFile.skipBytes(diff);

        // create a byte stream to contain the input object
        ObjectBytesInputStream bytes = new ObjectBytesInputStream(data);

        // create a stream to translate the object
        SerialObjectInputStream istrm = new SerialObjectInputStream(bytes);

        // return the object
        Object record = istrm.readObject();
        
        // return record; apply filter is necessary
        if (m_filter == null)
            return record;
        else
            return m_filter.filterDecode(record);
    }

    /**
     * Deletes the record in the current file position.
     * 
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    // delete the current object record
    public void delete() throws IOException
    {
        // assume beginning of record, get pointer
        long pos = m_dataFile.getFilePointer();

        // is this record already deleted?
        byte deleted = m_dataFile.readByte();

        if (deleted == IS_DELETED)
            return;

        // back up and write new marker
        m_dataFile.seek(pos);
        m_dataFile.writeByte(IS_DELETED);

        // skip length value
        m_dataFile.skipBytes(4 + 4);

        // write next-deleted marker from header
        m_dataFile.writeLong(m_firstDeleted);

        // now change file header
        m_firstDeleted = pos;
        m_dataFile.seek(0);
        m_dataFile.writeLong(m_firstDeleted);
    }

    /**
     * Sets the current file pointer to the first byte beyond the file header.
     * 
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    // go to beginning of file
    public void rewind() throws IOException
    {
        m_dataFile.seek(8);
    }

    /**
     * Skips the record at the current file position, moving to the next record in sequence.
     * 
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     */
    // skip over an object record
    public void skip() throws IOException
    {
        // skip over parts of record header
        m_dataFile.skipBytes(1 + 4);

        // read record length
        int reclen = m_dataFile.readInt();

        // skip to beginning of next record
        m_dataFile.skipBytes(8 + reclen);
    }

    /**
     * Compacts a database file by copying it to a new file and deleting the old one.
     * The records will be copied in sequence, and the callback function is called for
     * each record as it is rewritten. The callback function is responsible for rebuilding
     * any indexes or other referneces to the records, which will have new file positions 
     * after compaction.
     * 
     * @param   callback                called for each record rewritten to the compacted file
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     */
    // compact file by removing deleted records
    public void compact(ObjectDatabaseCallback callback) throws IOException, ClassNotFoundException
    {
        // generate temporary file name
        File tempName = new File("TMP" + System.currentTimeMillis());

        // create the temporary file
        RandomAccessFile newFile = new RandomAccessFile(tempName,"rw");

        // write file header
        newFile.writeLong(DEL_LIST_END);

        // read and write records
        while (true)
        {
            try
            {
                byte deleted = m_dataFile.readByte();
                int  dataLen = m_dataFile.readInt();
                int  recLen  = m_dataFile.readInt();

                if (deleted == IS_DELETED)
                {
                    // skip to next record
                    m_dataFile.skipBytes(8 + recLen);
                }
                else
                {
                    // skip deleted marker
                    m_dataFile.skipBytes(8);

                    // create array for input data
                    byte [] data = new byte [dataLen];

                    // read raw data
                    m_dataFile.readFully(data);

                    // skip over any waste space
                    int diff = recLen - dataLen;

                    if (diff > 0)
                        m_dataFile.skipBytes(diff);

                    // remember where this record starts
                    long pos = newFile.getFilePointer();

                    // write new header to output file
                    newFile.writeByte(IS_ACTIVE);
                    newFile.writeInt(dataLen);
                    newFile.writeInt(dataLen);
                    newFile.writeLong(DEL_LIST_END);

                    // write raw data
                    newFile.write(data);

                    // create a byte stream to contain the input object
                    ObjectBytesInputStream bytes = new ObjectBytesInputStream(data);

                    // create a stream to translate the object
                    SerialObjectInputStream istrm = new SerialObjectInputStream(bytes);

                    // notify the callback method
                    if (callback != null)
                        callback.compactNotify(pos,istrm.readObject());
                }
            }
            catch ( EOFException eof )
            {
                break;
            }
        }

        // close and delete the old file
        m_dataFile.close();

        File goAway = new File(m_fileName);

        goAway.delete();

        // rename the new file
        newFile.close();

        tempName.renameTo(new File(m_fileName));

        // now we use the new file
        m_dataFile = new RandomAccessFile(m_fileName,"rw");
        constructorHelper();
    }
    
    /**
     * Assigns a filter object to translated, encrypt, compress, or otherwise manipulate
     * objects as they are written and read.
     * 
     * @param   filter      assigned filter
     */
    // filter methods
    public void assignFilter(RecordFilter filter)
    {
        m_filter = filter;
    }

    public final FileDescriptor getFD() throws IOException
    {
        return m_dataFile.getFD();
    }

    public long getFilePointer() throws IOException
    {
        return m_dataFile.getFilePointer();
    }

    public void seek(long pos) throws IOException
    {
        m_dataFile.seek(pos);
    }

    public long length() throws IOException
    {
        return m_dataFile.length();
    }

    public void close() throws IOException
    {
        m_dataFile.close();
    }

    protected ObjectInputStream createObjectInputStream(InputStream in)
        throws IOException
    {
        return new ObjectInputStream(in);
    }

    protected ObjectOutputStream createObjectOutputStream(OutputStream out)
        throws IOException
    {
        return new ObjectOutputStream(out);
    }
}

