//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  ObjectIndex.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/** * <code>ObjectIndex</code> defines the concept of an <code>index</code> that associates <code>key</code>
 * values with references to serializable objects. Concrete subclasses of <code>ObjectIndex</code>
 * assume that an object's <code>position</code> (as given by a <code>long</code> value) is constant,
 * regardless of the actual meaning of that position. In general, the position represents the * location (file pointer) of a <code>Serializable</code> object in an <code>IndexedObjectDatabase</code>. * <p> * Use the <code>IndexedObjectDatabase.attachIndex</code> method to attach an <code>ObjectIndex</code> * to a database. A database updates all attached indexes when it writes or removes objects * from its file.
 * <p>
 * @author	Scott Robert Ladd
 * @see		com.coyotegulch.jisp.ObjectIndex
 * @see		com.coyotegulch.jisp.IndexedObjectDatabase
 */
public interface ObjectIndex
{
    /**
     * Insert a key into the database, associating a record position with the given key.
     * Throws a <code>DuplicateKey</code> exception if the key already exists.    
     * <p>     
     * @param   key         A key identifying the record to be read.
     * @param   pos         File position of record associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  DuplicateKey            when inserting a duplicate key into an index that does not
     *                                  support duplicates
     * @see     com.coyotegulch.jisp.DuplicateKey
     * @see     com.coyotegulch.jisp.KeyObject
     */
    void insertKey(KeyObject key, long pos) throws IOException, ClassNotFoundException;
    
    /**
     * Replace the reference pos for the given key. Throws a <code>KeyNotFound</code>
     * exception if the requested key does not exist in the index.
     *
     * @param   key         A key identifying the record to be read.
     * @param   pos         File position of record associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    void replaceKey(KeyObject key, long pos) throws IOException, ClassNotFoundException;
                                                           
    /**
     * If the key already exists, replace the reference pos for the given key. Otherwise, insert
     * a key into the database, associating a record position with the given key.
     * 
     * @param   key         A key identifying the record to be read.
     * @param   pos         File position of record associated with key
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @see     com.coyotegulch.jisp.KeyObject
     */
    void storeKey(KeyObject key, long pos) throws IOException, ClassNotFoundException;
    
    /**
     * Find the position of the object associated with a given key.
     * @param   key         A key identifying the record to be read.
     * @return  The position of the record associated with <code>key</code>.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    long findKey(KeyObject key) throws IOException, ClassNotFoundException;
    
    /**
     * Removes the given key from the index.
     *
     * @param   key         A key identifying the record to be read.
     * @throws  IOException             when an I/O exception is thrown by an underlying java.io.* class
     * @throws  ClassNotFoundException  for a casting error, usually when a persistent object or index does
     *                                  match the expected type
     * @throws  KeyNotFound             when the specified key can not be found in the index
     * @see     com.coyotegulch.jisp.KeyNotFound
     * @see     com.coyotegulch.jisp.KeyObject
     */
    void removeKey(KeyObject key) throws IOException, ClassNotFoundException;
}
