//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  SerialObjectOutputStream.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//
//  Based on an idea submitted by Rickard Oberg.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;
import java.util.*;

class SerialObjectOutputStream extends OutputStream
{
    private static final int DEFAULT_BLOCK_SIZE = 8192;

    private byte[]     m_buffer;
    private LinkedList m_buffers;
    private int        m_index;
    private int        m_size;
    private int        m_blockSize;
    private boolean    m_closed;

    public SerialObjectOutputStream()
    {
        this(DEFAULT_BLOCK_SIZE);
    }

    public SerialObjectOutputStream(int aSize)
    {
        m_blockSize = aSize;
        m_buffer    = new byte[m_blockSize];
    }

    // Public
    public void writeTo(OutputStream out) throws IOException
    {
        // Check if we have a list of m_buffers
        if (m_buffers != null)
        {
            Iterator enum = m_buffers.iterator();

            while (enum.hasNext())
            {
                byte[] bytes = (byte[]) enum.next();
                out.write(bytes, 0, m_blockSize);
            }
        }

        // write the internal m_buffer directly
        out.write(m_buffer, 0, m_index);
    }

    public void writeTo(RandomAccessFile out) throws IOException
    {
        // Check if we have a list of m_buffers
        if (m_buffers != null)
        {
            Iterator enum = m_buffers.iterator();
            while (enum.hasNext())
            {
                byte[] bytes = (byte[]) enum.next();
                out.write(bytes, 0, m_blockSize);
            }
        }

        // write the internal m_buffer directly
        out.write(m_buffer, 0, m_index);
    }

    public int getSize()
    {
        return m_size+m_index;
    }

    public byte[] toByteArray()
    {
        byte[] data = new byte[getSize()];
 
       // Check if we have a list of m_buffers
       int pos = 0;
       if (m_buffers != null)
       {
           Iterator enum = m_buffers.iterator();
           while (enum.hasNext())
           {
               byte[] bytes = (byte[]) enum.next();
               System.arraycopy(bytes,0,data,pos,m_blockSize);
               pos+=m_blockSize;
           }
       }

       // write the internal m_buffer directly
       System.arraycopy(m_buffer, 0, data, pos, m_index);

       return data;
    }

    public void write(int datum) throws IOException
    {
        if (m_closed)
            throw new IOException("Stream m_closed");
        else
        {
            if (m_index == m_blockSize)
            {
                // Create new m_buffer and store current in linked list
                if (m_buffers == null)
                    m_buffers = new LinkedList();

                m_buffers.addLast(m_buffer);

                m_buffer = new byte[m_blockSize];
                m_size += m_index;
                m_index = 0;
            }

            // store the byte
            m_buffer[m_index++] = (byte) datum;
        }
    }

    public void write(byte[] data, int offset, int length) throws IOException
    {
       if (data == null)
           throw new NullPointerException();
       else
       {
           if ((offset < 0) || (offset + length > data.length) || (length < 0))
               throw new IndexOutOfBoundsException();
           else
           {
               if (m_closed)
                   throw new IOException("Stream m_closed");
               else
               {
                   if (m_index + length > m_blockSize)
                   {
                       // Write byte by byte
                       for (int i = 0; i < length; i++)
                            write(data[offset+i]);
                    }
                    else
                    {
                        // Copy in the subarray
                        System.arraycopy(data, offset, m_buffer, m_index, length);
                        m_index += length;
                    }
                }
            }
        }
    }

    public void close()
    {
        m_closed = true;
    }
}
