//-----------------------------------------------------------------------
//	Java Indexed Serialization Package
//  ----------------------------------
//  StringKey.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any legal purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

package com.coyotegulch.jisp;

import java.io.*;

/**
 * <code>StringKey</code> defines a 16-character fixed-length text key.
 * <p>
 * @author	Scott Robert Ladd
 * @see		com.coyotegulch.jisp.KeyObject
 * @see		com.coyotegulch.jisp.ObjectIndex
 */
public class StringKey extends KeyObject
{
    static final long serialVersionUID = -6894793231339165076L;
    
    // maximum key length
    protected static final int KEY_LENGTH = 16;
    
    // working storage
    private String m_key;
    
    /**
     * Creates a new <code>StringKey</code> using <code>key_value</code>.
     * <p>
     * @param   key_value   The text of the newly-created key.
     */    
    public StringKey(String key_value)
    {
        if (key_value.length() > KEY_LENGTH)
            m_key = key_value.substring(KEY_LENGTH);
        else
            m_key = new String(key_value);
    }
    
    /**
     * Creates a blank (null) <code>StringKey</code>.
     */    
    public StringKey()
    {
        m_key = new String();
//        m_key = new String("                ");
//                            1234567890123456
    }
    
    /**
     * Compares the invoking <code>StringKey</code> to <code>key</code>, returning one of the
     * <code>KEY_*</code> constants based on the relationship of the two keys.
     *
     * @param   key The <code>StringKey</code> value to be compared against the invoking key.
     * @return  one of the <code>KEY_*</code> constants based on the relationship of the two keys.     
     */
    public int compareTo(KeyObject key)
    {
        if (key instanceof StringKey)
        {
            int comp = m_key.trim().compareTo(((StringKey)key).m_key.trim());
    
            if (comp == 0)
                return KEY_EQUAL;
            else
                if (comp < 0)
                    return KEY_LESS;
                else
                    return KEY_MORE;
        }
        else
            return KEY_ERROR;
    }
    
    /**
     * Returns a "null" (i.e., blank) <code>StringKey</code>, to be used as a space-holder in certain kinds of
     * <code>ObjectIndexes</code>.
     *
     * @return  a "null" (i.e., blank) <code>StringKey</code> constant.
     */
    public KeyObject makeNullKey()
    {
        return new StringKey();
    }
    
    /**
     * <code>StringKey</code> implements the <code>writeExternal</code> method to save its
     * contents.
     */
    public void writeExternal(ObjectOutput out) throws IOException
    {
        String outKey;
        
        int x = m_key.length();
        
        if (m_key.length() < KEY_LENGTH)
        {
            outKey = new String(m_key);
                
            for (int n = 0; n < (KEY_LENGTH - m_key.length()); ++n)
                outKey += ' ';
        }
        else
            outKey = m_key.substring(0,KEY_LENGTH);
             
        out.writeUTF(outKey);
    }
    
    /**
     * <code>StringKey</code> implements the readExternal method to restore its contents.
     */
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException
    {
        // read data
        m_key = in.readUTF();
    }

    /**
     * Returns the <code>String</code> representation of a <code>StringKey</code>.
     *
     * @return  <code>String</code> representation of a <code>StringKey</code>.
     */
    public String toString()
    {
        return m_key;
    }
    
    /**
     * Returns a hash code value for the object, which is, in effect, the    
     * the sum of the characters in the string key value.    
     *
     * @return  hash code value for this object.
     * @see     java.lang.Object    
     * @see     com.coyotegulch.jisp.HashIndex
     */
    public int hashCode()
    {
        int hash = 0;
        
        int max = (m_key.length() < KEY_LENGTH) ? m_key.length() : KEY_LENGTH;
        
        for (int i = 0; i < max; ++i)
            hash += (int)m_key.charAt(i);
                   
        return hash;
    }
}
